import logging

import numpy.random

from ..events import job_context
from ..events import workflow_context
from ..progress import TextProgress
from .examples.tasks.sumlist import SumList


def test_no_progress_stdout(capsys):
    task = SumList(
        inputs={"list": numpy.random.random(10000)},
    )
    assert not task.done
    task.execute()
    stdout = capsys.readouterr()
    assert len(stdout.out) == 0
    assert task.done


def test_text_progress_stdout(capsys):
    task = SumList(
        inputs={"list": numpy.random.random(10000)},
        progress=TextProgress(name="SumList"),
    )
    assert not task.done
    task.execute()
    stdout = capsys.readouterr()
    assert len(stdout.out) > 0
    assert stdout.out.count("DONE") > 0
    assert task.done


def test_no_progress_events(caplog):
    with caplog.at_level(logging.INFO):
        with job_context(None) as execinfo:
            with workflow_context(execinfo) as execinfo:
                task = SumList(
                    inputs={"list": numpy.random.random(10000)},
                    node_id="node_id",
                    execinfo=execinfo,
                )
                task.execute()
    assert len(caplog.records) == 6


def test_text_progress_events(caplog):
    with caplog.at_level(logging.INFO):
        with job_context(None) as execinfo:
            with workflow_context(execinfo) as execinfo:
                task = SumList(
                    inputs={"list": numpy.random.random(10000)},
                    progress=TextProgress(name="SumList"),
                    node_id="node_id",
                    execinfo=execinfo,
                )
                task.execute()
    nprogress = 100
    assert len(caplog.records) == (nprogress + 6)
    progress = [
        record.progress for record in caplog.records if record.type == "progress"
    ]
    assert len(progress) == nprogress
    assert progress == list(range(1, nprogress + 1))



import inspect
import logging
import textwrap
from dataclasses import dataclass
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Type

import networkx
from ewoksutils import import_utils

from ..dynamictask import get_dynamically_task_class
from ..missing_data import MISSING_DATA
from ..missing_data import is_missing_data
from ..model import BaseInputModel
from ..node import NodeIdType
from ..task import Task
from .taskgraph import TaskGraph

_logger = logging.getLogger(__name__)


@dataclass
class NodeInput:
    id: NodeIdType
    label: Optional[str]
    task_identifier: str
    name: str
    value: Any
    required: bool
    description: Optional[str]
    examples: Optional[List[Any]]
    import_error: Optional[Exception]

    @property
    def has_value(self) -> bool:
        return not is_missing_data(self.value)

    @property
    def required_without_value(self) -> bool:
        return self.required and not self.has_value


def graph_inputs(graph: TaskGraph) -> List[NodeInput]:
    """
    Return a list of workflow inputs. These are all the task
    inputs that are not connected to task outputs from previous
    nodes in the workflow.
    """
    node_inputs = _get_node_inputs(graph.graph)

    task_identifiers = set(node_input.task_identifier for node_input in node_inputs)
    short_task_ids = _shorten_task_identifiers(task_identifiers)
    for node_input in node_inputs:
        node_input.task_identifier = short_task_ids[node_input.task_identifier]

    return node_inputs


def graph_inputs_as_table(
    graph: TaskGraph, column_widths: Optional[Dict[str, Optional[int]]] = None
) -> Tuple[List[str], List[List[str]], Dict[str, str], List[str]]:
    """
    Return table of workflow input parameters.
    """
    node_inputs = graph_inputs(graph)
    column_names, rows, footnotes = _graph_inputs_to_table(
        node_inputs, column_widths=column_widths
    )
    metadata = {}
    if graph.graph_id:
        metadata["id"] = graph.graph_id
    if graph.graph_label:
        metadata["description"] = graph.graph_label
    return column_names, rows, metadata, footnotes


def _graph_inputs_to_table(
    node_inputs: List[NodeInput],
    column_widths: Optional[Dict[str, Optional[int]]] = None,
) -> Tuple[List[str], List[List[str]], List[str]]:
    """
    Convert a list of workflow inputs to a table with string values.
    """
    if column_widths is None:
        column_widths = {
            "name": None,
            "value": 30,
            "description": 40,
            "examples": 30,
            "task_identifier": None,
            "id": None,
            "label": None,
        }

    # Column names
    column_names = [s.replace("_", " ").capitalize() for s in column_widths]

    # Highlight required inputs without a value
    highlighted = [
        node_input for node_input in node_inputs if node_input.required_without_value
    ]
    node_inputs = highlighted + [
        node_input
        for node_input in node_inputs
        if not node_input.required_without_value
    ]

    # Generate table
    rows = []
    has_import_error = False
    for node_input in node_inputs:
        row = []
        for column_name, width in column_widths.items():
            value = _get_row_value(node_input, column_name)
            str_val = _row_value_as_string(value, width)
            row.append(str_val)
        rows.append(row)

    # Remove empty columns
    columns = list(zip(*rows))
    non_empty_column_indices = [
        i for i, col in enumerate(columns) if any(cell.strip() for cell in col)
    ]
    column_names = [column_names[i] for i in non_empty_column_indices]
    rows = [[row[i] for i in non_empty_column_indices] for row in rows]

    # Footnotes
    footnotes = []
    has_required_without_value = any(
        node_input.required_without_value for node_input in node_inputs
    )
    if has_required_without_value:
        footnotes = ["⁽*⁾ Value is required for execution."]
    has_import_error = any(node_input.import_error for node_input in node_inputs)
    if has_import_error:
        footnotes.append(
            "⁽†⁾ Information from workflow only (task cannot be imported)."
        )

    return column_names, rows, footnotes


def _get_row_value(node_input: NodeInput, column_name: str) -> Any:
    value = getattr(node_input, column_name)
    if column_name == "name":
        if node_input.required_without_value:
            return f"{value}⁽*⁾"
        if node_input.import_error:
            return f"{value}⁽†⁾"
    if column_name == "value":
        return repr(value)
    if column_name == "examples" and value:
        return list(map(repr, value))
    return value


def _row_value_as_string(value: Any, width: Optional[int]) -> str:
    if value is None:
        str_val = ""
    elif isinstance(value, str):
        str_val = value
    elif isinstance(value, list):
        if width:
            str_val = _wrap_bullet_list(value, width)
            width = None
        else:
            str_val = "• " + "\n• ".join(value)
    else:
        str_val = str(value)

    if width:
        return "\n".join(textwrap.wrap(str_val, width=width))
    return str_val


def _wrap_bullet_list(items: List[str], width: int) -> str:
    wrapper = textwrap.TextWrapper(
        width=width, initial_indent="• ", subsequent_indent="  "
    )
    return "\n".join(wrapper.fill(str(item)) for item in items)


def _get_node_inputs(graph: networkx.DiGraph) -> List[NodeInput]:
    """
    Return all the task inputs that are not connected to task
    outputs from previous nodes in the workflow.
    """
    all_node_inputs = []
    for node_id, node_attrs in graph.nodes.items():
        node_inputs = _get_all_node_inputs(node_id, node_attrs)
        connected_input_names = _get_connected_input_names(graph, node_id)
        all_node_inputs += [
            node_input
            for node_input in node_inputs
            if node_input.name not in connected_input_names
        ]
    return all_node_inputs


def _get_connected_input_names(
    graph: networkx.DiGraph, node_id: NodeIdType
) -> Set[str]:
    """
    Return all input parameter names that are connected to an output from a previous task.
    """
    connected_input_names = set()
    for predecessor_id in graph.predecessors(node_id):
        link_attrs = graph.get_edge_data(predecessor_id, node_id)

        if not link_attrs:
            continue

        data_mappings = link_attrs.get("data_mapping", [])
        for mapping in data_mappings:
            target_input = mapping.get("target_input")
            if target_input:
                connected_input_names.add(target_input)

        map_all_data = link_attrs.get("map_all_data", False)
        if map_all_data:
            node_attrs = graph.nodes[predecessor_id]
            task_type = node_attrs["task_type"]
            task_identifier = node_attrs["task_identifier"]
            output_names = _get_all_task_output_names(task_type, task_identifier)
            connected_input_names.update(output_names)
    return connected_input_names


def _get_all_node_inputs(
    node_id: NodeIdType, node_attrs: Dict[str, Any]
) -> List[NodeInput]:
    """
    Return all the input parameters of a node.
    """
    task_type = node_attrs["task_type"]
    task_identifier = node_attrs["task_identifier"]
    default_inputs = node_attrs.get("default_inputs", [])
    default_input_map = {item["name"]: item.get("value") for item in default_inputs}

    node_attrs = {
        "id": node_id,
        "label": node_attrs.get("label", None),
        "task_identifier": task_identifier,
    }

    if task_type == "class":
        try:
            task_cls = import_utils.import_qualname(task_identifier)
        except Exception as import_error:
            _logger.warning(f"Cannot import {task_identifier!r}: {import_error}")
            node_input_iterator = _node_inputs_from_defaults(
                default_input_map, node_attrs, import_error
            )
        else:
            node_input_iterator = _node_inputs_from_class(
                task_cls, default_input_map, node_attrs
            )
    elif task_type == "generated":
        try:
            task_cls = get_dynamically_task_class(
                node_attrs.get("task_generator"), task_identifier
            )
        except Exception as import_error:
            _logger.warning(f"Cannot import {task_identifier!r}: {import_error}")
            node_input_iterator = _node_inputs_from_defaults(
                default_input_map, node_attrs, import_error
            )
        else:
            node_input_iterator = _node_inputs_from_class(
                task_cls, default_input_map, node_attrs
            )
    elif task_type in ("method", "ppfmethod"):
        try:
            task_method = import_utils.import_qualname(task_identifier)
        except Exception as import_error:
            _logger.warning(f"Cannot import {task_identifier!r}: {import_error}")
            node_input_iterator = _node_inputs_from_defaults(
                default_input_map, node_attrs, import_error
            )
        else:
            node_input_iterator = _node_inputs_from_method(
                task_method, default_input_map, node_attrs
            )
    else:
        _logger.warning(
            f"Task type {task_type!r} is not supported ({task_identifier!r}). Only using default values from the workflow."
        )
        import_error = TypeError(f"Cannot get inputs from task type {task_type!r}")
        node_input_iterator = _node_inputs_from_defaults(
            default_input_map, node_attrs, import_error
        )

    return list(node_input_iterator)


def _get_all_task_output_names(task_type: str, task_identifier: str) -> List[str]:
    """
    Return all the output parameter names of a task.
    """
    if task_type == "class":
        try:
            task_cls = import_utils.import_qualname(task_identifier)
        except Exception:
            return []
        else:
            return _task_output_names_from_class(task_cls)
    elif task_type == "method":
        return ["return_value"]
    else:
        return []


def _node_inputs_from_class(
    task_cls: Type[Task], default_input_map: Dict[str, Any], node_attrs: Dict[str, Any]
) -> Generator[NodeInput, None, None]:
    """
    Return all task input parameters based on a task class.
    """
    input_model = task_cls.input_model()
    if input_model:
        yield from _node_inputs_from_class_model(
            input_model, default_input_map, node_attrs
        )
    else:
        yield from _node_inputs_from_class_methods(
            task_cls, default_input_map, node_attrs
        )


def _node_inputs_from_class_model(
    input_model: BaseInputModel,
    default_input_map: Dict[str, Any],
    node_attrs: Dict[str, Any],
) -> Generator[NodeInput, None, None]:
    """
    Return all task input parameters based on an input model.
    """
    for name, field in input_model.model_fields.items():
        required = field.is_required()

        if name in default_input_map:
            # Default input overwrites model default (if any)
            value = default_input_map[name]
        else:
            try:
                default = field.get_default()
            except Exception:
                # Field has no default value
                value = MISSING_DATA
            else:
                # Field has a default value
                value = default

        yield NodeInput(
            **node_attrs,
            name=name,
            required=required,
            value=value,
            description=field.description,
            examples=field.examples,
            import_error=None,
        )


def _node_inputs_from_class_methods(
    task_cls: Type[Task], default_input_map: Dict[str, Any], node_attrs: Dict[str, Any]
) -> Generator[NodeInput, None, None]:
    """
    Return all task input parameters based on a task class.
    """
    input_names = [(name, True) for name in sorted(task_cls.required_input_names())]
    input_names += [(name, False) for name in sorted(task_cls.optional_input_names())]

    for name, required in input_names:
        value = default_input_map.get(name, MISSING_DATA)
        yield NodeInput(
            **node_attrs,
            name=name,
            required=required,
            value=value,
            description=None,
            examples=None,
            import_error=None,
        )


def _node_inputs_from_method(
    task_method: Callable, default_input_map: Dict[str, Any], node_attrs: Dict[str, Any]
) -> Generator[NodeInput, None, None]:
    """
    Return all task input parameters based on a task method.
    """
    sig = inspect.signature(task_method)
    for name, param in sig.parameters.items():
        if param.kind in (param.kind.VAR_POSITIONAL, param.kind.VAR_KEYWORD):
            continue

        required = param.default is inspect.Parameter.empty
        if name in default_input_map:
            # Default input overwrites parameter default (if any)
            value = default_input_map[name]
        elif required:
            # Parameter has no default value
            value = MISSING_DATA
        else:
            # Parameter has a default value
            value = param.default

        yield NodeInput(
            **node_attrs,
            name=name,
            required=required,
            value=value,
            description=None,
            examples=None,
            import_error=None,
        )


def _task_output_names_from_class(task_cls) -> List[str]:
    return sorted(task_cls.output_names())


def _node_inputs_from_defaults(
    default_input_map: Dict[str, Any],
    node_attrs: Dict[str, Any],
    import_error: Exception,
) -> Generator[NodeInput, None, None]:
    for name, value in default_input_map.items():
        yield NodeInput(
            **node_attrs,
            name=name,
            required=True,
            value=value,
            description=None,
            examples=None,
            import_error=import_error,
        )


def _shorten_task_identifiers(task_identifiers: Sequence[str]) -> Dict[str, str]:
    """
    Return a mapping from full task identifiers to the shortest unique suffixes.
    """
    task_identifiers = set(task_identifiers)
    nunique = len(task_identifiers)

    all_reversed_parts = {
        tid: tuple(reversed(tid.split("."))) for tid in task_identifiers
    }
    reversed_parts = {
        tid: (tid_parts[0],) for tid, tid_parts in all_reversed_parts.items()
    }

    while True:
        all_parts = list(reversed_parts.values())
        nunique_current = len(set(all_parts))
        if nunique_current == nunique:
            break
        for tid, tid_parts in list(reversed_parts.items()):
            if all_parts.count(tid_parts) == 1:
                continue
            i = len(tid_parts)
            full_tid_parts = all_reversed_parts[tid]
            if i < len(full_tid_parts):
                reversed_parts[tid] = reversed_parts[tid] + (full_tid_parts[i],)

    return {pid: ".".join(reversed(parts)) for pid, parts in reversed_parts.items()}



import enum
import importlib
import json
import logging
import os
from collections.abc import Mapping
from pathlib import Path
from typing import Optional
from typing import Tuple
from typing import Union

import networkx
import yaml
from ewoksutils.deprecation_utils import deprecated
from ewoksutils.path_utils import makedirs_from_filename
from packaging.version import Version

from ..node import node_id_from_json
from .models import GraphSource
from .schema import normalize_schema_version

logger = logging.getLogger(__name__)

GraphRepresentation = enum.Enum(
    "GraphRepresentation", "json json_dict json_string json_module yaml test_core"
)

network_x_version = Version(networkx.__version__)


def dump(
    graph: networkx.DiGraph,
    destination: Optional[Union[str, Path]] = None,
    representation: Optional[Union[GraphRepresentation, str]] = None,
    **save_options,
) -> Union[str, Path, dict]:
    """From runtime to persistent representation"""
    if isinstance(representation, str):
        representation = GraphRepresentation.__members__[representation]
    if representation is None:
        if isinstance(destination, (str, Path)):
            filename = str(destination).lower()
            if filename.endswith(".json"):
                representation = GraphRepresentation.json
            elif filename.endswith((".yml", ".yaml")):
                representation = GraphRepresentation.yaml
        else:
            representation = GraphRepresentation.json_dict

    if representation == GraphRepresentation.json_dict:
        return _networkx_to_dict(graph)

    if representation == GraphRepresentation.json:
        if destination is None:
            raise TypeError("Destination should be specified when dumping to json")
        dictrepr = dump(graph)
        makedirs_from_filename(destination)
        save_options.setdefault("indent", 2)
        with open(destination, mode="w") as f:
            json.dump(dictrepr, f, **save_options)
        return destination

    if representation == GraphRepresentation.json_string:
        dictrepr = dump(graph)
        return json.dumps(dictrepr, **save_options)

    if representation == GraphRepresentation.yaml:
        if destination is None:
            raise TypeError("Destination should be specified when dumping to yaml")
        dictrepr = dump(graph)
        makedirs_from_filename(destination)
        with open(destination, mode="w") as f:
            yaml.dump(dictrepr, f, **save_options)
        return destination

    if representation == GraphRepresentation.json_module:
        if destination is None:
            raise TypeError("Destination should be specified when dumping to json")
        package, _, file = str(destination).rpartition(".")
        assert package, f"No package provided when saving graph to '{destination}'"
        destination = os.path.join(_package_path(package), f"{file}.json")
        return dump(
            graph, destination=destination, representation="json", **save_options
        )

    if representation == GraphRepresentation.test_core:
        raise TypeError("'test_core' workflows representations cannot be saved")

    raise TypeError(representation, type(representation))


def load(
    source: Optional[GraphSource] = None,
    representation: Optional[Union[GraphRepresentation, str]] = None,
    root_dir: Optional[Union[str, Path]] = None,
    root_module: Optional[str] = None,
) -> networkx.DiGraph:
    """From persistent to runtime representation"""
    if isinstance(representation, str):
        representation = GraphRepresentation.__members__[representation]

    if representation is None:
        if isinstance(source, Mapping):
            representation = GraphRepresentation.json_dict
        elif isinstance(source, (str, Path)):
            if isinstance(source, str) and "{" in source and "}" in source:
                representation = GraphRepresentation.json_string
            else:
                filename = str(source).lower()
                if filename.endswith(".json"):
                    representation = GraphRepresentation.json
                elif filename.endswith((".yml", ".yaml")):
                    representation = GraphRepresentation.yaml
                else:
                    source = _read_any_file(
                        source, root_dir=root_dir, root_module=root_module
                    )
                    representation = GraphRepresentation.json_dict

    if not source:
        graph = networkx.DiGraph()
    elif isinstance(source, networkx.Graph):
        graph = source
    elif hasattr(source, "graph") and isinstance(source.graph, networkx.Graph):
        graph = source.graph
    elif representation == GraphRepresentation.json_dict:
        graph = _dict_to_networkx(source)
    elif representation == GraphRepresentation.json:
        graph_dict = _read_json_file(source, root_dir=root_dir, root_module=root_module)
        graph = _dict_to_networkx(graph_dict)
    elif representation == GraphRepresentation.json_string:
        graph_dict = json_load(source)
        graph = _dict_to_networkx(graph_dict)
    elif representation == GraphRepresentation.yaml:
        graph_dict = _read_yaml_file(source, root_dir=root_dir, root_module=root_module)
        graph = _dict_to_networkx(graph_dict)
    elif representation == GraphRepresentation.json_module:
        package, _, source = source.rpartition(".")
        if package:
            source = os.path.join(_package_path(package), source)
        return load(
            source,
            representation="json",
            root_dir=root_dir,
            root_module=root_module,
        )
    elif representation == GraphRepresentation.test_core:
        from ..tests.examples.graphs import get_graph

        return load(get_graph(source)[0], representation="json_dict")
    else:
        raise TypeError(representation, type(representation))

    if not networkx.is_directed(graph):
        raise TypeError(graph, type(graph))

    return graph


def _read_json_file(
    filename: Union[str, Path],
    root_dir: Optional[Union[str, Path]] = None,
    root_module: Optional[str] = None,
) -> dict:
    filename = _find_graph_path(
        filename,
        root_dir=root_dir,
        root_module=root_module,
        possible_extensions=(".json",),
    )
    with open(filename, mode="r") as f:
        return json_load(f)


def _read_yaml_file(
    filename: str, root_dir: Optional[str] = None, root_module: Optional[str] = None
) -> dict:
    filename = _find_graph_path(
        filename,
        root_dir=root_dir,
        root_module=root_module,
        possible_extensions=(".yml", ".yaml"),
    )
    with open(filename, mode="r") as f:
        return _yaml_load(f)


def _read_any_file(
    filename: Union[str, Path],
    root_dir: Optional[Union[str, Path]] = None,
    root_module: Optional[str] = None,
) -> Optional[dict]:
    filename = _find_graph_path(
        filename,
        root_dir=root_dir,
        root_module=root_module,
        possible_extensions=(".json", ".yml", ".yaml"),
    )
    with open(filename, mode="r") as f:
        content = f.read()

    try:
        return json_load(content)
    except (json.JSONDecodeError, TypeError):
        pass

    try:
        return _yaml_load(content)
    except (yaml.YAMLError, TypeError):
        pass

    raise ValueError(f"File format of '{filename}' not supported")


def json_load(content) -> dict:
    if isinstance(content, str):
        result = json.loads(content, object_pairs_hook=_ewoks_jsonload_hook)
    else:
        result = json.load(content, object_pairs_hook=_ewoks_jsonload_hook)
    if not isinstance(result, Mapping):
        raise TypeError("graph must be a dictionary")
    return result


def _yaml_load(content) -> dict:
    result = yaml.load(content, yaml.Loader)
    if not isinstance(result, Mapping):
        raise TypeError("graph must be a dictionary")
    return result


def _ewoks_jsonload_hook_pair(item):
    key, value = item
    if key in (
        "source",
        "target",
        "sub_source",
        "sub_target",
        "id",
        "node",
        "sub_node",
    ):
        value = node_id_from_json(value)
    return key, value


@deprecated("Use 'json_load' instead")
def ewoks_jsonload_hook(items):
    return _ewoks_jsonload_hook(items)


def _ewoks_jsonload_hook(items):
    return dict(map(_ewoks_jsonload_hook_pair, items))


def _find_graph_path(
    path: Union[str, Path],
    root_dir: Optional[Union[str, Path]] = None,
    root_module: Optional[str] = None,
    possible_extensions: Tuple[str] = tuple(),
) -> str:
    """When the :code:`path` is relative, the parent directory is assumed to be
    (in order of priority):

    * :code:`root_dir`
    * :code:`root_module` directory
    * current working directory

    When :code:`path` is not found it tries to find the path with a different
    extension from :code:`possible_extensions`.

    :param path: could be a relative path, might have no extension
    :param root_dir: in case :code:`path` is relative
    :param root_module: in case :code:`root_module` is not provided
    :param possible_extensions: in case :code:`path` is not found
    :raises: FileNotFoundError
    """
    # Absolute path
    if not root_dir and root_module:
        root_dir = _package_path(root_module)
    if not os.path.isabs(path) and root_dir:
        path = os.path.join(root_dir, path)
    path = os.path.abspath(path)

    if os.path.exists(path):
        return path

    # Try different extensions
    root, _ = os.path.splitext(path)
    for new_ext in possible_extensions:
        new_full_path = root + new_ext
        if os.path.exists(new_full_path):
            return new_full_path

    raise FileNotFoundError(path)


def _package_path(package: str) -> str:
    package = importlib.import_module(package)
    return package.__path__[0]


def _dict_to_networkx(graph: dict) -> networkx.DiGraph:
    graph.setdefault("directed", True)
    graph.setdefault("nodes", list())
    graph.setdefault("links", list())
    graph.setdefault("graph", dict())

    if "id" not in graph["graph"]:
        logger.warning('Graph has no "id": use "notspecified"')
        graph["graph"]["id"] = "notspecified"
    normalize_schema_version(graph)

    if network_x_version < Version("3.4rc"):
        return networkx.readwrite.json_graph.node_link_graph(graph)
    else:
        return networkx.readwrite.json_graph.node_link_graph(graph, edges="links")


def _networkx_to_dict(graph: networkx.DiGraph) -> dict:
    if network_x_version < Version("3.4rc"):
        graph_dict = networkx.readwrite.json_graph.node_link_data(graph)
    else:
        graph_dict = networkx.readwrite.json_graph.node_link_data(graph, edges="links")

    # Remove fields that are not part of the Ewoks spec
    graph_dict.pop("directed")
    graph_dict.pop("multigraph")
    return graph_dict



import logging
import threading
import time
from pathlib import Path
from pprint import pformat
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional

import pytest
from ewoksutils.event_utils import FIELD_TYPES
from ewoksutils.import_utils import qualname
from ewoksutils.sqlite3_utils import connect
from ewoksutils.sqlite3_utils import select

from ..bindings import execute_graph
from ..events import cleanup as cleanup_events
from ..task import Task

logger = logging.getLogger(__name__)


@pytest.fixture
def sqlite_path(tmp_path) -> Generator[Path, None, None]:
    try:
        yield tmp_path
    finally:
        cleanup_events()


def test_succesfull_workfow(sqlite_path: Path):
    database = sqlite_path / "ewoks_events.db"
    run_succesfull_workfow(database, execute_graph)
    events = fetch_events(database, 10)
    assert_succesfull_workfow_events(events)


def test_failed_workfow(sqlite_path: Path):
    database = sqlite_path / "ewoks_events.db"
    run_failed_workfow(database, execute_graph)
    events = fetch_events(database, 8)
    assert_failed_workfow_events(events)


def test_changing_handlers(sqlite_path: Path):
    database1 = sqlite_path / "ewoks_events1.db"
    run_succesfull_workfow(database1, execute_graph, delay=0.2)
    events = fetch_events(database1, 10)
    assert_succesfull_workfow_events(events)

    size_before = database1.stat().st_size

    database2 = sqlite_path / "ewoks_events2.db"
    run_succesfull_workfow(database2, execute_graph, delay=0.2)
    events = fetch_events(database2, 10)
    assert_succesfull_workfow_events(events)

    size_after = database1.stat().st_size
    assert size_before == size_after


def test_changing_handlers_parallel(sqlite_path: Path, n_concurrent: int = 4):
    databases = [sqlite_path / f"ewoks_events_{i}.db" for i in range(n_concurrent)]

    threads = [
        threading.Thread(
            target=run_succesfull_workfow,
            args=(db, execute_graph),
            kwargs=dict(delay=0.2),
        )
        for db in databases
    ]

    _run_threads(threads)

    sizes = []
    for db in databases:
        events = fetch_events(db, 10)
        assert_succesfull_workfow_events(events)
        sizes.append(db.stat().st_size)

    assert len(set(sizes)) == 1


def _run_threads(threads):
    for t in threads:
        t.start()

    deadline = time.time() + 20

    for t in threads:
        remaining = deadline - time.time()
        if remaining <= 0:
            pytest.fail("Timeout while waiting for workflow threads to finish")

        t.join(timeout=remaining)

        if t.is_alive():
            pytest.fail(f"Workflow thread {t.name} did not finish within {deadline}s")


class MyTask(
    Task,
    input_names=["ctr"],
    optional_input_names=["error_msg", "delay"],
    output_names=["ctr"],
):
    def run(self):
        if self.inputs.error_msg:
            raise ValueError(self.inputs.error_msg)
        else:
            self.outputs.ctr = self.inputs.ctr + 1
        if self.inputs.delay:
            time.sleep(self.inputs.delay)


def run_succesfull_workfow(
    database: Path, execute_graph, delay: int = 0, **execute_options
):
    graph = {"id": "test_graph", "schema_version": "1.1"}
    nodes = [
        {
            "id": "node1",
            "task_type": "class",
            "task_identifier": qualname(MyTask),
            "default_inputs": [
                {"name": "ctr", "value": 0},
                {"name": "delay", "value": delay},
            ],
        },
        {
            "id": "node2",
            "task_type": "class",
            "task_identifier": qualname(MyTask),
            "default_inputs": [
                {"name": "ctr", "value": 0},
                {"name": "delay", "value": delay},
            ],
        },
        {
            "id": "node3",
            "task_type": "class",
            "task_identifier": qualname(MyTask),
            "default_inputs": [
                {"name": "ctr", "value": 0},
                {"name": "delay", "value": delay},
            ],
        },
    ]
    links = [
        {
            "source": "node1",
            "target": "node2",
            "data_mapping": [{"source_output": "ctr", "target_input": "ctr"}],
        },
        {
            "source": "node2",
            "target": "node3",
            "data_mapping": [{"source_output": "ctr", "target_input": "ctr"}],
        },
    ]
    taskgraph = {"graph": graph, "nodes": nodes, "links": links}
    _execute_graph(database, taskgraph, execute_graph, **execute_options)


def assert_succesfull_workfow_events(events):
    expected = [
        {"context": "job", "node_id": None, "type": "start"},
        {"context": "workflow", "node_id": None, "type": "start"},
        {"context": "node", "node_id": "node1", "type": "start"},
        {"context": "node", "node_id": "node1", "type": "end"},
        {"context": "node", "node_id": "node2", "type": "start"},
        {"context": "node", "node_id": "node2", "type": "end"},
        {"context": "node", "node_id": "node3", "type": "start"},
        {"context": "node", "node_id": "node3", "type": "end"},
        {"context": "workflow", "node_id": None, "type": "end"},
        {"context": "job", "node_id": None, "type": "end"},
    ]
    captured = [
        {k: event[k] for k in ("context", "node_id", "type")} for event in events
    ]
    _assert_events(expected, captured)


def run_failed_workfow(
    database, execute_graph, delay: int = 0, **execute_options
) -> None:
    graph = {"id": "test_graph", "schema_version": "1.1"}
    nodes = [
        {
            "id": "node1",
            "task_type": "class",
            "task_identifier": qualname(MyTask),
            "default_inputs": [
                {"name": "ctr", "value": 0},
                {"name": "delay", "value": delay},
            ],
        },
        {
            "id": "node2",
            "task_type": "class",
            "task_identifier": qualname(MyTask),
            "default_inputs": [
                {"name": "ctr", "value": 0},
                {"name": "error_msg", "value": "abc"},
                {"name": "delay", "value": delay},
            ],
        },
        {
            "id": "node3",
            "task_type": "class",
            "task_identifier": qualname(MyTask),
            "default_inputs": [
                {"name": "ctr", "value": 0},
                {"name": "delay", "value": delay},
            ],
        },
    ]
    links = [
        {
            "source": "node1",
            "target": "node2",
            "data_mapping": [{"source_output": "ctr", "target_input": "ctr"}],
        },
        {
            "source": "node2",
            "target": "node3",
            "data_mapping": [{"source_output": "ctr", "target_input": "ctr"}],
        },
    ]
    graph = {"graph": graph, "nodes": nodes, "links": links}
    _execute_graph(database, graph, execute_graph, **execute_options)


def assert_failed_workfow_events(events):
    err_msg = "Execution failed for ewoks task 'node2' (id: 'node2', task: 'ewokscore.tests.test_workflow_events.MyTask'): abc"

    expected = [
        {
            "context": "job",
            "node_id": None,
            "type": "start",
            "error_message": None,
        },
        {
            "context": "workflow",
            "node_id": None,
            "type": "start",
            "error_message": None,
        },
        {
            "context": "node",
            "node_id": "node1",
            "type": "start",
            "error_message": None,
        },
        {
            "context": "node",
            "node_id": "node1",
            "type": "end",
            "error_message": None,
        },
        {
            "context": "node",
            "node_id": "node2",
            "type": "start",
            "error_message": None,
        },
        {
            "context": "node",
            "node_id": "node2",
            "type": "end",
            "error_message": "abc",
        },
        {
            "context": "workflow",
            "node_id": None,
            "type": "end",
            "error_message": err_msg,
        },
        {
            "context": "job",
            "node_id": None,
            "type": "end",
            "error_message": err_msg,
        },
    ]
    captured = [
        {k: event[k] for k in ("context", "node_id", "type", "error_message")}
        for event in events
    ]
    _assert_events(expected, captured)


def _execute_graph(database, graph, execute_graph, **execute_options) -> None:
    execinfo = execute_options.setdefault("execinfo", dict())
    handlers = execinfo.setdefault("handlers", list())
    handlers.append(
        {
            "class": "ewokscore.events.handlers.Sqlite3EwoksEventHandler",
            "arguments": [{"name": "uri", "value": str(database)}],
        }
    )
    try:
        execute_graph(graph, **execute_options)
    except RuntimeError:
        pass


def _assert_events(expected, captured):
    missing = list()
    unexpected = list(captured)
    for event in expected:
        try:
            unexpected.remove(event)
        except ValueError:
            missing.append(event)
    if missing or unexpected:
        raise AssertionError(
            f"ewoks events not as expected\nmissing:\n{pformat(missing)}\nunexpected:\n{unexpected}"
        )


def fetch_events(database: Path, nevents: int) -> List[Dict[str, Optional[str]]]:
    """Events are handled asynchronously so wait until we have the required `nevents`
    up to 3 seconds.
    """
    exception = None
    events = list()
    for _ in range(30):
        try:
            with connect(database) as conn:
                events = list(select(conn, "ewoks_events", field_types=FIELD_TYPES))

            if len(events) != nevents:
                raise RuntimeError(f"{len(events)} ewoks events instead of {nevents}")
            return events
        except Exception as e:
            exception = e
            time.sleep(0.1)
    if exception:
        logger.error(exception)
    return events



from typing import List

import pytest

from .. import task_discovery
from .conftest import expected_tasks


@pytest.mark.parametrize("task_type", ["class", "method", "ppfmethod", None])
def test_discover_tasks_from_one_module(task_type):
    expected = expected_tasks("ewokscore.tests.discover.module1", task_type)

    tasks = task_discovery.discover_tasks_from_modules(
        "ewokscore.tests.discover.module1", task_type=task_type
    )
    assert_tasks(tasks, expected)
    assert len(tasks) == len(expected)


@pytest.mark.parametrize("task_type", ["class", "method", "ppfmethod", None])
def test_discover_tasks_from_module_pattern(task_type):
    expected = expected_tasks(task_type=task_type)

    tasks = task_discovery.discover_tasks_from_modules(
        "ewokscore.tests.discover.*", task_type=task_type
    )
    assert_tasks(tasks, expected)
    assert len(tasks) == len(expected)


def test_all_tasks_discovery():
    expected: List[task_discovery.TaskDict] = [
        {
            "category": "ewokscore",
            "optional_input_names": ["b", "delay"],
            "output_names": ["result", "too_small"],
            "required_input_names": ["a"],
            "task_identifier": "ewokscore.tests.examples.tasks.condsumtask.CondSumTask",
            "task_type": "class",
            "description": "Check whether a value is too small",
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
        {
            "category": "ewokscore",
            "optional_input_names": ["a", "b", "raise_error"],
            "output_names": ["result"],
            "required_input_names": [],
            "task_identifier": "ewokscore.tests.examples.tasks.errorsumtask.ErrorSumTask",
            "task_type": "class",
            "description": "Add two number with intentional exception",
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
        {
            "category": "ewokscore",
            "optional_input_names": [],
            "output_names": [],
            "required_input_names": [],
            "task_identifier": "ewokscore.tests.examples.tasks.nooutputtask.NoOutputTask",
            "task_type": "class",
            "description": "A task without outputs",
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
        {
            "category": "ewokscore",
            "optional_input_names": ["delay"],
            "output_names": ["sum"],
            "required_input_names": ["list"],
            "task_identifier": "ewokscore.tests.examples.tasks.sumlist.SumList",
            "task_type": "class",
            "description": "Add items from a list",
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
        {
            "category": "ewokscore",
            "optional_input_names": ["b", "delay"],
            "output_names": ["result"],
            "required_input_names": ["a"],
            "task_identifier": "ewokscore.tests.examples.tasks.sumtask.SumTask",
            "task_type": "class",
            "description": "Add two numbers with a delay",
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
        {
            "category": "ewokscore",
            "task_identifier": "ewokscore.tests.examples.tasks.addfunc.addfunc",
            "task_type": "method",
            "required_input_names": ["arg"],
            "optional_input_names": [],
            "output_names": ["return_value"],
            "description": "Add 1 to the first argument",
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
        {
            "category": "ewokscore",
            "task_identifier": "ewokscore.tests.examples.tasks.simplemethods.add",
            "task_type": "method",
            "required_input_names": [],
            "optional_input_names": [],
            "output_names": ["return_value"],
            "description": "Sum objects and add 1",
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
        {
            "category": "ewokscore",
            "task_identifier": "ewokscore.tests.examples.tasks.simplemethods.append",
            "task_type": "method",
            "required_input_names": [],
            "optional_input_names": [],
            "output_names": ["return_value"],
            "description": "Return positional arguments as a tuple",
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
    ]

    tasks = task_discovery.discover_all_tasks()
    assert_tasks(tasks, expected)

    for task_type in ("class", "method", "ppfmethod"):
        tasks = task_discovery.discover_all_tasks(task_type=task_type)
        assert_tasks(
            tasks, [task for task in expected if task["task_type"] == task_type]
        )


def _find_task(tasks, identifier, task_type):
    for task in tasks:
        if task["task_identifier"] == identifier and task["task_type"] == task_type:
            return task

    raise ValueError(f"Task {identifier} and type {task_type} not found")


def assert_tasks(tasks, expected):
    for task in tasks:
        if task["category"] != "ewokscore":
            continue
        expected_task = _find_task(expected, task["task_identifier"], task["task_type"])
        for key, value in task.items():
            assert expected_task[key] == value



from dataclasses import dataclass
from typing import Union

import pytest
from pydantic import BaseModel
from pydantic import field_validator

from ..missing_data import MISSING_DATA
from ..missing_data import MissingData
from ..missing_data import is_missing_data
from ..model import BaseInputModel
from ..task import Task
from ..task import TaskInputError
from ..variable import Variable
from .examples.tasks.sumtask import SumTask


class User(BaseInputModel):
    id: int
    name: str = "Jane Doe"


class PassThroughTask(Task, input_model=User, output_names=["result"]):
    def run(self):
        self.outputs.result = self.get_input_values()


def test_error_if_input_model_does_not_derive_from_base_model():
    class WrongBaseModelUser(BaseModel):
        id: int
        name: str = "Jane Doe"

    with pytest.raises(
        TypeError,
        match=r"input_model should be a subclass of ewokscore.model.BaseInputModel",
    ):

        class WrongPassThroughTask(Task, input_model=WrongBaseModelUser):
            pass


def test_error_if_input_model_used_with_input_names():
    with pytest.raises(TypeError, match="input_model cannot be used with input_names"):

        class WrongPassThroughTask(
            Task, input_model=User, input_names=["age"], output_names=["result"]
        ):
            pass


def test_validation():
    with pytest.raises(TaskInputError, match=r"Missing inputs.+\['id'\]"):
        PassThroughTask(inputs={})

    task = PassThroughTask(inputs={"id": "wrong type"})
    with pytest.raises(RuntimeError, match=r"id(\s*)Input should be a valid integer"):
        task.execute()


def test_default_value():
    task = PassThroughTask(inputs={"id": 5})
    task.execute()
    assert task.get_input_values() == {"id": 5, "name": "Jane Doe"}


def test_wrapped_value(tmp_path):
    varinfo = {"root_uri": str(tmp_path / "task_results")}
    variable = Variable(5, varinfo=varinfo)
    variable.dump()
    varinfo = {"root_uri": str(tmp_path)}

    task = PassThroughTask(inputs={"id": variable})
    task.execute()

    task = PassThroughTask(inputs={"id": variable.uhash}, varinfo=varinfo)
    task.execute()

    task = PassThroughTask(inputs={"id": variable.data_uri})
    task.execute()

    task = PassThroughTask(inputs={"id": variable.data_proxy})
    task.execute()


def test_wrapped_wrong_value(tmp_path):
    varinfo = {"root_uri": str(tmp_path / "task_results")}
    variable = Variable("wrong type", varinfo=varinfo)
    variable.dump()
    varinfo = {"root_uri": str(tmp_path)}

    task = PassThroughTask(inputs={"id": variable})
    with pytest.raises(RuntimeError, match=r"id(\s*)Input should be a valid integer"):
        task.execute()

    task = PassThroughTask(inputs={"id": variable.uhash}, varinfo=varinfo)
    with pytest.raises(RuntimeError, match=r"id(\s*)Input should be a valid integer"):
        task.execute()

    task = PassThroughTask(inputs={"id": variable.data_uri})
    with pytest.raises(RuntimeError, match=r"id(\s*)Input should be a valid integer"):
        task.execute()

    task = PassThroughTask(inputs={"id": variable.data_proxy})
    with pytest.raises(RuntimeError, match=r"id(\s*)Input should be a valid integer"):
        task.execute()


def test_run():
    task = PassThroughTask(inputs={"id": 5, "name": "Smith"})
    task.execute()
    assert task.outputs["result"] == {"id": 5, "name": "Smith"}


def test_error_on_subclass_with_wrong_submodel():
    class Car(BaseInputModel):
        wheels: int

    with pytest.raises(
        TypeError,
        match="Input model (.*) from task subclass must be a subclass of the original task input model",
    ):

        class PassThroughCarTask(PassThroughTask, input_model=Car):
            pass


def test_error_on_subclass_with_input_names():
    with pytest.raises(
        TypeError,
        match="Cannot use input_names or optional_input_names",
    ):

        class ChildPassThroughTask(PassThroughTask, input_names=["age"]):
            pass


def test_error_on_subclass_with_input_model_if_input_names():
    with pytest.raises(
        TypeError,
        match="Cannot use input_model",
    ):

        class ChildPassThroughTask(SumTask, input_model=User):
            pass


def test_subclass_with_no_change():
    class ChildPassThroughTask(PassThroughTask):
        pass

    task = ChildPassThroughTask(inputs={"id": 5, "name": "Smith"})
    task.execute()
    assert task.outputs["result"] == {"id": 5, "name": "Smith"}


class SuperUser(User):
    age: int


class PassThroughSubTask(PassThroughTask, input_model=SuperUser):
    pass


def test_subclass_validation():
    with pytest.raises(TaskInputError, match=r"Missing inputs.+\['age'\]"):
        PassThroughSubTask(inputs={"id": 5})


def test_subclass():
    task = PassThroughSubTask(inputs={"id": 5, "age": 18})
    task.execute()
    assert task.outputs["result"] == {"id": 5, "name": "Jane Doe", "age": 18}


def test_missing_data():
    class RegularTask(Task, input_names=["one"], optional_input_names=["two"]):
        pass

    class Model(BaseInputModel):
        one: int
        two: Union[int, MissingData] = MISSING_DATA

    class ModelTask(Task, input_model=Model):
        pass

    regular_task = RegularTask(inputs={"one": 1})
    model_task = ModelTask(inputs={"one": 1})
    assert (
        model_task.get_input_values() == regular_task.get_input_values() == {"one": 1}
    )
    assert is_missing_data(model_task.get_input_value("two"))
    assert is_missing_data(regular_task.get_input_value("two"))
    assert model_task.missing_inputs["two"]
    assert regular_task.missing_inputs["two"]


class UserWithTypeCoercion(User):
    age: int

    @field_validator("age", mode="before")
    def coerce_age(cls, value):
        if isinstance(value, float):
            return int(value + 0.5)
        if not isinstance(value, int):
            return -1
        return value


class TaskWithTypeCoercion(Task, input_model=UserWithTypeCoercion):
    def run(self):
        pass


def test_input_type_coercion():
    task = TaskWithTypeCoercion(inputs={"id": 5, "age": 18})
    task.execute()
    assert task.get_input_values() == {"id": 5, "name": "Jane Doe", "age": 18}

    task = TaskWithTypeCoercion(inputs={"id": 5, "age": 18.1})
    task.execute()
    assert task.get_input_values() == {"id": 5, "name": "Jane Doe", "age": 18}

    task = TaskWithTypeCoercion(inputs={"id": 5, "age": "wrong type"})
    task.execute()
    assert task.get_input_values() == {"id": 5, "name": "Jane Doe", "age": -1}


def test_wrapped_type_coercion(tmp_path):
    varinfo = {"root_uri": str(tmp_path / "task_results")}
    variable = Variable(18.1, varinfo=varinfo)
    variable.dump()
    varinfo = {"root_uri": str(tmp_path)}

    coerced_input_values = {"id": 5, "name": "Jane Doe", "age": 18}

    inputs = {"id": 5, "name": "Jane Doe", "age": 18}
    task = TaskWithTypeCoercion(inputs=inputs, varinfo=varinfo)
    task.execute()
    assert task.get_input_values() == coerced_input_values
    coerced_variable_uhash = task.input_variables["age"].uhash
    # Includes input name "age" in hashing so not equal
    assert coerced_variable_uhash != variable.uhash

    inputs = {"id": 5, "name": "Jane Doe", "age": 18.1}
    task = TaskWithTypeCoercion(inputs=inputs, varinfo=varinfo)
    task.execute()
    assert task.get_input_values() == coerced_input_values
    uhash = task.input_variables["age"].uhash
    assert uhash == coerced_variable_uhash

    inputs = {"id": 5, "name": "Jane Doe", "age": Variable(18, varinfo=varinfo)}
    task = TaskWithTypeCoercion(inputs=inputs, varinfo=varinfo)
    task.execute()
    assert task.get_input_values() == coerced_input_values
    uhash = task.input_variables["age"].uhash
    assert uhash == Variable(18, varinfo=varinfo).uhash

    expected_uhash = uhash
    inputs = {"id": 5, "name": "Jane Doe", "age": Variable(18.1, varinfo=varinfo)}
    task = TaskWithTypeCoercion(inputs=inputs, varinfo=varinfo)
    task.execute()
    assert task.get_input_values() == coerced_input_values
    uhash = task.input_variables["age"].uhash
    assert uhash == expected_uhash

    uhash_before = variable.uhash
    inputs = {"id": 5, "name": "Jane Doe", "age": variable}
    task = TaskWithTypeCoercion(inputs=inputs, varinfo=varinfo)
    task.execute()
    assert task.get_input_values() == coerced_input_values
    input_variable = task.input_variables["age"]
    assert input_variable.uhash == variable.uhash
    assert variable.uhash != uhash_before

    # Reset variable since it got modified in-memory in the previous task execution
    varinfo = {"root_uri": str(tmp_path / "task_results")}
    variable = Variable(18.1, varinfo=varinfo)
    # No need for dump: was only modified in memory
    varinfo = {"root_uri": str(tmp_path)}
    fixed_uhash = variable.uhash

    for reference in [fixed_uhash, variable.data_uri, variable.data_proxy]:
        inputs = {"id": 5, "name": "Jane Doe", "age": reference}
        task = TaskWithTypeCoercion(inputs=inputs, varinfo=varinfo)
        task.execute()
        assert task.get_input_values() == coerced_input_values
        input_variable = task.input_variables["age"]
        assert input_variable.uhash == fixed_uhash
        assert variable.uhash == fixed_uhash


def test_dataclass_field_stays_dataclass():
    @dataclass
    class Address:
        city: str
        street: str

    class Inputs(BaseInputModel):
        address: Address

    class CheckAddress(Task, input_model=Inputs):
        def run(self):
            address = self.inputs.address
            assert isinstance(address, Address)

    task = CheckAddress(
        inputs={"address": Address(city="Grenoble", street="Jean-Jaurès")}
    )
    task.execute()



import networkx


def v0_update(graph: networkx.DiGraph) -> None:
    """Outdated version"""
    raise RuntimeError("not supported")


def from_v1_0_to_v1_1(graph: networkx.DiGraph) -> None:
    """This version does not have the requirements field."""
    graph.graph["schema_version"] = "1.1"



import hashlib
import random
from collections.abc import Iterable
from collections.abc import Mapping
from collections.abc import Set
from typing import Any
from typing import Optional
from typing import Type
from typing import Union

import numpy
from ewoksutils.import_utils import qualname

from . import missing_data


def classhashdata(cls: Type) -> bytes:
    return qualname(cls).encode()


def multitype_sorted(sequence: Iterable, key=None) -> list:
    try:
        return sorted(sequence, key=key)
    except TypeError:
        pass
    if key is None:

        def key(item):
            return item

    adict = dict()
    for item in sequence:
        typename = type(key(item)).__name__
        adict.setdefault(typename, list()).append(item)

    return [
        item
        for _, items in sorted(adict.items(), key=lambda tpl: tpl[0])
        for item in sorted(items, key=key)
    ]


class UniversalHash:
    def __init__(self, hexdigest: Union[str, bytes]):
        if isinstance(hexdigest, bytes):
            hexdigest = hexdigest.decode()
        if not isinstance(hexdigest, str):
            raise TypeError(hexdigest, type(hexdigest))
        self._hexdigest = hexdigest

    def __hash__(self):
        # make it python hashable (to use in sets and dict keys)
        return hash(self._hexdigest)

    def __repr__(self):
        return "UniversalHash('{}')".format(self)

    def __str__(self):
        return self._hexdigest

    def __eq__(self, other):
        return str(self) == str(other)

    def __lt__(self, other):
        return str(self) < str(other)


def uhash(value, _hash=None) -> UniversalHash:
    """Universial hash (as opposed to python's `hash`)."""
    # Avoid using python's hash!
    bdigest = _hash is None
    if bdigest:
        _hash = hashlib.sha256()
    _hash.update(classhashdata(type(value)))
    if value is None:
        pass
    elif isinstance(value, HasUhash):
        _hash.update(repr(value.uhash).encode())
    elif isinstance(value, UniversalHash):
        _hash.update(repr(value).encode())
    elif isinstance(value, bytes):
        _hash.update(value)
    elif isinstance(value, str):
        _hash.update(value.encode())
    elif isinstance(value, int):
        _hash.update(hex(value).encode())
    elif isinstance(value, float):
        _hash.update(value.hex().encode())
    elif isinstance(value, (numpy.ndarray, numpy.number)):
        _hash.update(value.tobytes())
    elif isinstance(value, Mapping):
        lst = multitype_sorted(value.items(), key=lambda item: item[0])
        if lst:
            keys, values = zip(*lst)
        else:
            keys = values = list()
        uhash(keys, _hash=_hash)
        uhash(values, _hash=_hash)
    elif isinstance(value, Set):
        values = multitype_sorted(value)
        uhash(values, _hash=_hash)
    elif isinstance(value, Iterable):
        # Ordered
        for v in value:
            uhash(v, _hash=_hash)
    else:
        # TODO: register custom types
        raise TypeError(f"cannot uhash {value} (type: {type(value)})")
    if bdigest:
        return UniversalHash(_hash.hexdigest())


class HasUhash:
    @property
    def uhash(self) -> Optional[UniversalHash]:
        raise NotImplementedError

    def __hash__(self):
        # make it python hashable (to use in sets and dict keys)
        uhash = self.uhash
        if uhash is None:
            return hash(id(self))
        else:
            return hash(uhash)

    def __eq__(self, other):
        if isinstance(other, HasUhash):
            uhash = other.uhash
        elif isinstance(other, UniversalHash):
            uhash = other
        else:
            raise TypeError(other, type(other))
        return self.uhash == uhash

    def _get_repr_data(self) -> dict:
        data = dict()
        uhash = self.uhash
        if uhash is None:
            data["uhash"] = None
        else:
            data["uhash"] = repr(str(uhash))
        return data

    def __repr__(self):
        data = self._get_repr_data()
        if data:
            sdata = ", ".join([f"{k}={v}" for k, v in data.items()])
            return f"{super().__repr__()}({sdata})"
        else:
            return super().__repr__()

    def __str__(self):
        data = self._get_repr_data()
        if data:
            sdata = ", ".join([f"{k}={v}" for k, v in data.items()])
            return f"{qualname(type(self))}({sdata})"
        else:
            return qualname(type(self))


PreUhashTypes = Union[str, bytes, UniversalHash, HasUhash]


class UniversalHashable(HasUhash):
    """The universal hash of an instance of this class is based on:

     * pre-uhash
     * instance nonce (if any)

    The universal hash is equal to the pre-hash when an instance nonce is not provided.

    The pre-uhash is either provided or based on:

     * data
     * class nonce (class qualifier name, class version, superclass nonce)
    """

    __CLASS_NONCE = None
    __VERSION = None
    MISSING_DATA = missing_data.MISSING_DATA

    def __init__(
        self,
        pre_uhash: Optional[PreUhashTypes] = None,
        instance_nonce: Optional[Any] = None,
    ):
        self.set_uhash_init(pre_uhash=pre_uhash, instance_nonce=instance_nonce)

    def __init_subclass__(subcls, version=None, **kwargs):
        super().__init_subclass__(**kwargs)
        supercls_data = subcls.class_nonce()
        subcls.__VERSION = version
        subcls_data = subcls.class_nonce_data()
        subcls.__CLASS_NONCE = str(uhash((subcls_data, supercls_data)))

    def set_uhash_init(
        self,
        pre_uhash: Optional[PreUhashTypes] = None,
        instance_nonce: Optional[Any] = None,
    ):
        self.__set_pre_uhash(pre_uhash)
        self.__original_pre_uhash = self.__pre_uhash
        self.__instance_nonce = instance_nonce
        self.__original__instance_nonce = instance_nonce

    def get_uhash_init(self, serialize=False):
        pre_uhash = self.__original_pre_uhash
        if serialize:
            if isinstance(pre_uhash, HasUhash):
                pre_uhash = str(pre_uhash.uhash)
            elif isinstance(pre_uhash, UniversalHash):
                pre_uhash = str(pre_uhash)
        return {
            "pre_uhash": pre_uhash,
            "instance_nonce": self.__original__instance_nonce,
        }

    def __set_pre_uhash(self, pre_uhash):
        if pre_uhash is None:
            self.__pre_uhash = None
        elif isinstance(pre_uhash, (str, bytes)):
            self.__pre_uhash = UniversalHash(pre_uhash)
        elif isinstance(pre_uhash, (UniversalHash, HasUhash)):
            self.__pre_uhash = pre_uhash
        else:
            self.__pre_uhash = uhash(pre_uhash)

    @classmethod
    def class_nonce(cls):
        return cls.__CLASS_NONCE

    @classmethod
    def class_nonce_data(cls):
        return qualname(cls), cls.__VERSION

    def instance_nonce(self):
        return self.__instance_nonce

    def fix_uhash(self):
        """Fix the uhash when it is derived from the uhash data."""
        if self.__pre_uhash is not None:
            return
        keep, self.__instance_nonce = self.__instance_nonce, None
        try:
            pre_uhash = self.uhash
        finally:
            self.__instance_nonce = keep
        self.__set_pre_uhash(pre_uhash)

    def undo_fix_uhash(self):
        self.__pre_uhash = self.__original_pre_uhash

    def cleanup_references(self):
        """Remove all references to other hashables.
        Side effect: fixes the uhash when it depends on another hashable.
        """
        if isinstance(self.__pre_uhash, HasUhash):
            pre_uhash = self.__pre_uhash.uhash
            self.__pre_uhash = pre_uhash
            self.__original_pre_uhash = pre_uhash

    @property
    def uhash(self) -> Optional[UniversalHash]:
        _uhash = self.__pre_uhash
        if _uhash is None:
            data = self._uhash_data()
            if missing_data.is_missing_data(data):
                return None
            cnonce = self.class_nonce()
            inonce = self.instance_nonce()
            if inonce is None:
                return uhash((data, cnonce))
            else:
                return uhash((data, cnonce, inonce))
        else:
            if isinstance(_uhash, HasUhash):
                _uhash = _uhash.uhash
                if _uhash is None:
                    return None
            inonce = self.instance_nonce()
            if inonce is None:
                return _uhash
            else:
                return uhash((_uhash, inonce))

    def _uhash_data(self):
        return self.MISSING_DATA

    def uhash_randomize(self):
        self.__instance_nonce = random.randint(-1e100, 1e100)

    def undo_randomize(self):
        self.__instance_nonce = self.__original__instance_nonce



from . import graph


@graph
def self_trigger():
    graph = {"id": "self_trigger", "label": "self_trigger", "schema_version": "1.1"}

    task = "ewokscore.tests.examples.tasks.condsumtask.CondSumTask"
    nodes = [
        {
            "id": "task1",
            "default_inputs": [{"name": "a", "value": 1}],
            "task_type": "class",
            "task_identifier": task,
            "force_start_node": True,
        },
        {
            "id": "task2",
            "default_inputs": [{"name": "a", "value": 1}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task3",
            "task_type": "class",
            "task_identifier": task,
        },
    ]
    links = [
        {
            "source": "task1",
            "target": "task1",
            "data_mapping": [{"source_output": "result", "target_input": "b"}],
            "conditions": [{"source_output": "too_small", "value": True}],
        },
        {
            "source": "task1",
            "target": "task3",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task2",
            "target": "task3",
            "data_mapping": [{"source_output": "result", "target_input": "b"}],
        },
    ]

    expected = {
        "task3": {"too_small": False, "result": 11},
    }

    taskgraph = {"graph": graph, "links": links, "nodes": nodes}

    return taskgraph, expected



import sys
from typing import Tuple

if sys.version_info < (3, 9):
    from importlib.metadata import EntryPoint

    from importlib_metadata import entry_points as _entry_points

    def entry_points(group: str) -> Tuple[EntryPoint]:
        return tuple(_entry_points(group=group))

elif sys.version_info < (3, 10):
    from importlib.metadata import EntryPoint
    from importlib.metadata import entry_points as _entry_points

    def entry_points(group: str) -> Tuple[EntryPoint]:
        return _entry_points().get(group, ())

else:
    from importlib.metadata import EntryPoint
    from importlib.metadata import entry_points as _entry_points

    def entry_points(group: str) -> Tuple[EntryPoint]:
        return tuple(_entry_points(group=group))



import importlib
import pkgutil
from copy import deepcopy
from functools import wraps

_ALL_GRAPHS = None


def _discover_graphs():
    if _ALL_GRAPHS is None:
        for pkginfo in pkgutil.walk_packages(__path__, __name__ + "."):
            importlib.import_module(pkginfo.name)


def graph_names():
    _discover_graphs()
    return _ALL_GRAPHS.keys()


def get_graph(name) -> tuple:
    _discover_graphs()
    if name not in _ALL_GRAPHS:
        raise ValueError(f"{name!r} is not a known test graph")
    return deepcopy(_ALL_GRAPHS[name])


def graph(graph_method):
    global _ALL_GRAPHS
    name = graph_method.__name__

    @wraps(graph_method)
    def wrapper():
        g, result = graph_method()
        attrs = g.setdefault("graph", dict())
        assert attrs.get("id") == name
        assert attrs.get("label") == name
        assert attrs.get("schema_version") == "1.1"
        return g, result

    if _ALL_GRAPHS is None:
        _ALL_GRAPHS = dict()
    _ALL_GRAPHS[name] = wrapper()
    return wrapper



from typing import Callable
from typing import Optional
from typing import Tuple

import networkx
from packaging.version import parse as parse_version


class SchemaMetadata:
    """
    Metadata associated to each schema version:

        - ewokscore bounds: Versions of ewokscore that support this schema version
        - update method: A function that updates this schema version to the next one
    """

    def __init__(
        self,
        ewokscore_bounds: Tuple[str, Optional[str]],
        update_method: Optional[Callable[[networkx.DiGraph], None]],
    ) -> None:
        lower, upper = ewokscore_bounds
        if upper:
            self.ewokscore_bounds = parse_version(lower), parse_version(upper)
        else:
            self.ewokscore_bounds = parse_version(lower), None
        self.update_method = update_method



import itertools
from copy import deepcopy
from typing import Any
from typing import Iterator
from typing import Optional
from typing import Tuple
from typing import Union

import networkx

from ..node import flatten_node_id
from ..utils import dict_merge

NodeIdType = Union[str, Tuple[str, Any]]  # Any is NodeIdType


def _append_subnode_id(node_id: NodeIdType, sub_node_id: str) -> NodeIdType:
    if isinstance(node_id, tuple):
        parent, child = node_id
        return parent, _append_subnode_id(child, sub_node_id)
    else:
        return node_id, sub_node_id


def _get_subgraph(node_id: NodeIdType, subgraphs: dict):
    if isinstance(node_id, str):
        return subgraphs.get(node_id)
    subgraph_id, subnode_id = node_id
    try:
        subgraph = subgraphs[subgraph_id]
    except KeyError:
        raise ValueError(node_id, f"{repr(subgraph_id)} is not a subgraph")
    flat_subnode_id = flatten_node_id(subnode_id)
    n = len(flat_subnode_id)
    for node_id in subgraph.graph.nodes:
        flat_node_id = flatten_node_id(node_id)
        nflatid = len(flat_node_id)
        if flat_node_id == flat_subnode_id:
            return None  # a task node
        if nflatid > n and flat_node_id[:n] == flat_subnode_id:
            return subgraph  # a graph node
    raise ValueError(
        f"{subnode_id} is not a node or subgraph of subgraph {repr(subgraph_id)}",
    )


def _alias_to_node_id(alias_attrs: dict) -> NodeIdType:
    sub_node = alias_attrs.get("sub_node", None)
    if sub_node:
        return alias_attrs["node"], sub_node
    else:
        return alias_attrs["node"]


def _resolve_node_aliases(
    node_id: NodeIdType, graph_attrs: dict, input_nodes: bool
) -> Iterator[Tuple[NodeIdType, dict]]:
    if input_nodes:
        aliases = graph_attrs.get("input_nodes", list())
    else:
        aliases = graph_attrs.get("output_nodes", list())
    aliases = [alias_attrs for alias_attrs in aliases if alias_attrs["id"] == node_id]
    if aliases:
        for alias_attrs in aliases:
            sub_node_id = _alias_to_node_id(alias_attrs)
            link_attributes = alias_attrs.get("link_attributes", dict())
            yield sub_node_id, link_attributes
    else:
        yield node_id, dict()


def _resolve_all_node_aliases(
    graph_attrs: dict, input_nodes: bool
) -> Iterator[NodeIdType]:
    if input_nodes:
        aliases = graph_attrs.get("input_nodes", list())
    else:
        aliases = graph_attrs.get("output_nodes", list())
    for alias_attrs in aliases:
        yield _alias_to_node_id(alias_attrs)


def _get_subnode_ids(
    node_id: NodeIdType, link_attrs_subgraph_keys: dict, subgraphs: dict, source: bool
) -> Iterator[Tuple[NodeIdType, Optional[dict]]]:
    if source:
        key = "sub_source"
    else:
        key = "sub_target"
    subgraph = _get_subgraph(node_id, subgraphs)
    if subgraph is None:
        # node_id is not a subgraph
        if key in link_attrs_subgraph_keys:
            raise ValueError(
                f"'{node_id}' is not a graph so 'sub_source' should not be specified"
            )
        yield node_id, None
    else:
        # node_id is a subgraph
        try:
            sub_node_id = link_attrs_subgraph_keys[key]
        except KeyError:
            raise ValueError(
                f"the '{key}' attribute to specify a node in subgraph '{node_id}' is missing"
            ) from None
        for sub_node_id, link_attributes in _resolve_node_aliases(
            sub_node_id, subgraph.graph.graph, input_nodes=not source
        ):
            new_node_id = _append_subnode_id(node_id, sub_node_id)
            yield new_node_id, link_attributes


def _get_subnode_attributes(
    node_id: NodeIdType, subgraphs: dict, graph_node_attrs: dict
) -> Iterator[Tuple[NodeIdType, dict]]:
    """Update all input node attributes of the subgraph with the graph node attributes from the super graph"""
    transfer_attributes = {
        "default_inputs",
        "force_start_node",
        "conditions_else_value",
        "default_error_node",
    }
    node_attrs = {k: v for k, v in graph_node_attrs.items() if k in transfer_attributes}
    if not node_attrs:
        return
    subgraph = _get_subgraph(node_id, subgraphs)
    if subgraph is None:
        # node_id is not a subgraph
        return
    # node_id is a subgraph
    for sub_node_id in _resolve_all_node_aliases(
        subgraph.graph.graph, input_nodes=True
    ):
        new_node_id = _append_subnode_id(node_id, sub_node_id)
        yield new_node_id, node_attrs


def _get_subnode_links(
    source_id: NodeIdType,
    target_id: NodeIdType,
    link_attrs_subgraph_keys: dict,
    subgraphs: dict,
) -> Iterator[Tuple[NodeIdType, NodeIdType, dict, bool]]:
    sources = list(
        _get_subnode_ids(source_id, link_attrs_subgraph_keys, subgraphs, source=True)
    )
    targets = list(
        _get_subnode_ids(target_id, link_attrs_subgraph_keys, subgraphs, source=False)
    )
    for sub_source, source_link_attrs in sources:
        for sub_target, target_link_attrs in targets:
            if source_link_attrs:
                link_attrs = source_link_attrs
            else:
                link_attrs = dict()
            if target_link_attrs:
                link_attrs.update(target_link_attrs)
            target_is_graph = target_link_attrs is not None
            yield sub_source, sub_target, link_attrs, target_is_graph


def _replace_aliases(
    graph: networkx.DiGraph, subgraphs: dict, input_nodes: bool
) -> dict:
    if input_nodes:
        aliases = graph.graph.get("input_nodes")
        if not aliases:
            return
        source = False
        key = "sub_target"
    else:
        aliases = graph.graph.get("output_nodes")
        if not aliases:
            return
        source = True
        key = "sub_source"

    new_aliases = list()
    for alias_attrs in aliases:
        node_id = alias_attrs.get("node")
        if node_id is None:
            continue
        sub_node = alias_attrs.pop("sub_node", None)
        if sub_node:
            node_id = node_id, sub_node
        if not isinstance(node_id, tuple):
            new_aliases.append(alias_attrs)
            continue
        parent, child = node_id
        original_alias_attrs = alias_attrs
        for node_id, link_attrs in _get_subnode_ids(
            parent, {key: child}, subgraphs=subgraphs, source=source
        ):
            alias_attrs = deepcopy(original_alias_attrs)
            if link_attrs:
                link_attrs.update(alias_attrs.get("link_attributes", dict()))
                alias_attrs["link_attributes"] = link_attrs
            alias_attrs["node"] = node_id
            new_aliases.append(alias_attrs)

    if input_nodes:
        graph.graph["input_nodes"] = new_aliases
    else:
        graph.graph["output_nodes"] = new_aliases


def extract_graph_nodes(graph: networkx.DiGraph, subgraphs) -> Tuple[list, dict]:
    """Removes all graph nodes from `graph` and returns a list of edges
    between the nodes from `graph` and `subgraphs`.

    Nodes in sub-graphs are defines in the `link_attrs_subgraph_keys` link attribute.
    For example:

    .. code: python

        link_attrs = {
            "source": "subgraph1",
            "target": "subgraph2",
            "data_mapping": [{"source_output":"return_value", "target_input": 0}],
            "link_attrs_subgraph_keys": {
                "sub_source": ("subsubgraph", ("subsubsubgraph", "task2")),
                "sub_target": "task1",
            },
        }
    """
    edges = list()
    update_attrs = dict()
    graph_is_multi = graph.is_multigraph()
    for subgraph_id in subgraphs:
        # Nodes to be updated after flattening the graph
        for node_id, graph_node_attrs in _get_subnode_attributes(
            subgraph_id, subgraphs, graph.nodes[subgraph_id]
        ):
            if node_id in update_attrs:
                update_attrs[node_id].update(graph_node_attrs)
            else:
                update_attrs[node_id] = graph_node_attrs
        # Links to be made after flattening the graph
        it1 = (
            (source_id, subgraph_id) for source_id in graph.predecessors(subgraph_id)
        )
        it2 = ((subgraph_id, target_id) for target_id in graph.successors(subgraph_id))
        for source_id, target_id in itertools.chain(it1, it2):
            all_link_attrs = graph[source_id][target_id]
            if graph_is_multi:
                all_link_attrs = all_link_attrs.values()
            else:
                all_link_attrs = [all_link_attrs]
            for link_attrs in all_link_attrs:
                link_attrs_subgraph_keys = {
                    key: link_attrs.pop(key)
                    for key in ["sub_source", "sub_target", "sub_target_attributes"]
                    if key in link_attrs
                }
                if not link_attrs_subgraph_keys:
                    continue
                original_link_attrs = link_attrs
                links = _get_subnode_links(
                    source_id, target_id, link_attrs_subgraph_keys, subgraphs
                )
                for source, target, default_link_attrs, target_is_graph in links:
                    link_attrs = deepcopy(original_link_attrs)
                    if default_link_attrs:
                        link_attrs = {**default_link_attrs, **link_attrs}
                    sub_target_attributes = link_attrs_subgraph_keys.get(
                        "sub_target_attributes", None
                    )
                    if sub_target_attributes:
                        if not target_is_graph:
                            raise ValueError(
                                f"'{target_id}' is not a graph so 'sub_target_attributes' should not be specified"
                            )
                        if target in update_attrs:
                            update_attrs[target].update(sub_target_attributes)
                        else:
                            update_attrs[target] = sub_target_attributes
                    edges.append((source, target, link_attrs))

    _replace_aliases(graph, subgraphs, input_nodes=True)
    _replace_aliases(graph, subgraphs, input_nodes=False)
    graph.remove_nodes_from(subgraphs.keys())
    return edges, update_attrs


def add_subgraph_links(graph: networkx.DiGraph, edges: list, update_attrs: dict):
    # Output from extract_graph_nodes
    for source, target, _ in edges:
        if source not in graph.nodes:
            raise ValueError(
                f"Source node {repr(source)} of link |{repr(source)} -> {repr(target)}| does not exist"
            )
        if target not in graph.nodes:
            raise ValueError(
                f"Target node {repr(target)} of link |{repr(source)} -> {repr(target)}| does not exist"
            )
    graph.add_edges_from(edges)  # This adds missing nodes
    for node, attrs in update_attrs.items():
        if attrs:
            node_attrs = graph.nodes[node]
            dict_merge(node_attrs, attrs, overwrite=True)



# fmt: off
from ewoksutils.import_utils import instantiate_class as instantiate_handler  # noqa F401

from .base import *  # noqa F401
from .sqlite3 import Sqlite3EwoksEventHandler  # noqa F401

# fmt: on



from dataclasses import dataclass

import pytest
from pydantic import BaseModel
from pydantic import field_validator

from ..model import BaseOutputModel
from ..task import Task
from .examples.tasks.sumtask import SumTask


class User(BaseOutputModel):
    id: int
    name: str = "Jane Doe"


class PassThroughTask(
    Task, input_names=["id"], optional_input_names=["name"], output_model=User
):
    def run(self):
        self.outputs.id = self.inputs.id
        if not self.missing_inputs.name:
            self.outputs.name = self.inputs.name


def test_error_if_output_model_does_not_derive_from_base_model():
    class WrongBaseModelUser(BaseModel):
        id: int
        name: str = "Jane Doe"

    with pytest.raises(
        TypeError,
        match=r"output_model should be a subclass of ewokscore.model.BaseOutputModel",
    ):

        class WrongPassThroughTask(Task, output_model=WrongBaseModelUser):
            pass


def test_error_if_output_model_used_with_output_names():
    with pytest.raises(
        TypeError, match="output_model cannot be used with output_names"
    ):

        class WrongPassThroughTask(Task, output_model=User, output_names=["user"]):
            pass


def test_validation():
    task = PassThroughTask(inputs={"id": "wrong type"})
    with pytest.raises(RuntimeError, match=r"id(\s*)Input should be a valid integer"):
        task.execute()


def test_default_value():
    task = PassThroughTask(inputs={"id": 5})
    task.execute()
    assert task.get_output_values() == {"id": 5, "name": "Jane Doe"}


def test_run():
    task = PassThroughTask(inputs={"id": 5, "name": "Smith"})
    task.execute()
    assert task.get_output_values() == {"id": 5, "name": "Smith"}


def test_error_on_subclass_with_wrong_submodel():
    class Car(BaseOutputModel):
        wheels: int

    with pytest.raises(
        TypeError,
        match="Output model (.*) from task subclass must be a subclass of the original task output model",
    ):

        class PassThroughCarTask(PassThroughTask, output_model=Car):
            pass


def test_error_on_subclass_with_output_names():
    with pytest.raises(
        TypeError,
        match="Cannot use output_names",
    ):

        class ChildPassThroughTask(PassThroughTask, output_names=["age"]):
            pass


def test_error_on_subclass_with_output_model_if_output_names():
    with pytest.raises(
        TypeError,
        match="Cannot use output_model",
    ):

        class ChildPassThroughTask(SumTask, output_model=User):
            pass


def test_subclass_with_no_change():
    class ChildPassThroughTask(PassThroughTask):
        pass

    task = ChildPassThroughTask(inputs={"id": 5, "name": "Smith"})
    task.execute()
    assert task.get_output_values() == {"id": 5, "name": "Smith"}


class SuperUser(User):
    age: int


class PassThroughSubTask(
    PassThroughTask, optional_input_names=["age"], output_model=SuperUser
):
    def run(self):
        super().run()
        if not self.missing_inputs.age:
            self.outputs.age = self.inputs.age


def test_subclass_validation():
    task = PassThroughSubTask(inputs={"id": 5})
    with pytest.raises(RuntimeError, match=r"1 validation error for SuperUser\nage"):
        task.execute()


def test_subclass():
    task = PassThroughSubTask(inputs={"id": 5, "age": 18})
    task.execute()
    assert task.get_output_values() == {"id": 5, "name": "Jane Doe", "age": 18}


class UserWithTypeCoercion(User):
    age: int

    @field_validator("age", mode="before")
    def coerce_age(cls, value):
        if isinstance(value, float):
            return int(value + 0.5)
        if not isinstance(value, int):
            return -1
        return value


class TaskWithTypeCoercion(PassThroughTask, output_model=UserWithTypeCoercion):

    def run(self):
        super().run()
        if not self.missing_inputs.age:
            self.outputs.age = self.inputs.age


def test_output_type_coercion():
    task = TaskWithTypeCoercion(inputs={"id": 5, "age": 18})
    task.execute()
    assert task.get_output_values() == {"id": 5, "name": "Jane Doe", "age": 18}

    task = TaskWithTypeCoercion(inputs={"id": 5, "age": 18.1})
    task.execute()
    assert task.get_output_values() == {"id": 5, "name": "Jane Doe", "age": 18}

    task = TaskWithTypeCoercion(inputs={"id": 5, "age": "wrong type"})
    task.execute()
    assert task.get_output_values() == {"id": 5, "name": "Jane Doe", "age": -1}


def test_wrapped_type_coercion(tmp_path):
    varinfo = {"root_uri": str(tmp_path / "task_results")}
    task = TaskWithTypeCoercion(inputs={"id": 5, "age": 18}, varinfo=varinfo)
    task.execute()
    coerced_output_values = {"id": 5, "name": "Jane Doe", "age": 18}
    assert task.get_output_values() == coerced_output_values


def test_dataclass_field_stays_dataclass():
    @dataclass
    class Address:
        city: str
        street: str

    class Outputs(BaseOutputModel):
        address: Address

    class CheckAddress(Task, input_names=["address"], output_model=Outputs):
        def run(self):
            self.outputs.address = self.inputs.address

    address = Address(city="Grenoble", street="Jean-Jaurès")
    task = CheckAddress(inputs={"address": address})
    task.execute()
    assert isinstance(task.get_output_value("address"), Address)



from typing import List

import pytest
from ipykernel.kernelspec import install as install_kernel
from jupyter_client.kernelspec import KernelSpecManager
from packaging.version import parse as parse_version

from ..engine import CoreWorkflowEngine
from ..graph.schema import SchemaMetadata
from ..graph.schema import get_versions
from ..task_discovery import TaskDict


@pytest.fixture
def varinfo(tmp_path):
    yield {"root_uri": str(tmp_path)}


@pytest.fixture(scope="session")
def testkernel():
    m = KernelSpecManager()
    kernel_name = "pytest_kernel"
    install_kernel(kernel_name=kernel_name, user=True)
    yield kernel_name
    m.remove_kernel_spec(kernel_name)


@pytest.fixture
def use_test_schema_versions(monkeypatch):
    from ..graph import schema

    def no_update(graph):
        pass

    def backward_update(graph):
        graph.graph["schema_version"] = "0.1"

    def update_from_v0_2_to_1_0(graph):
        graph.graph["schema_version"] = "1.0"

    def get_test_versions():
        return {
            parse_version("0.1"): SchemaMetadata(("0.1.0-rc", None), no_update),
            parse_version("0.2"): SchemaMetadata(
                ("0.1.0-rc", None), update_from_v0_2_to_1_0
            ),
            parse_version("0.3"): SchemaMetadata(("0.1.0-rc", None), backward_update),
            **get_versions(),
        }

    monkeypatch.setattr(schema, "get_versions", get_test_versions)


def expected_tasks(module=None, task_type=None):
    CLASS_TASKS = [
        {
            "task_type": "class",
            "task_identifier": "ewokscore.tests.discover.module1.MyTask1",
            "required_input_names": ["a"],
            "optional_input_names": ["b"],
            "output_names": ["result"],
            "category": "ewokscore",
            "description": "Test 1",
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
        {
            "task_type": "class",
            "task_identifier": "ewokscore.tests.discover.module1.MyTask2",
            "required_input_names": ["a"],
            "optional_input_names": ["b"],
            "output_names": ["result"],
            "category": "ewokscore",
            "description": None,
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
        {
            "task_type": "class",
            "task_identifier": "ewokscore.tests.discover.module1.MyTask3",
            "required_input_names": [],
            "optional_input_names": [],
            "output_names": [],
            "category": "ewokscore",
            "description": None,
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 2,
        },
        {
            "task_type": "class",
            "task_identifier": "ewokscore.tests.discover.module2.MyTask3",
            "required_input_names": ["c", "z"],
            "optional_input_names": ["d", "x"],
            "output_names": ["error", "result"],
            "category": "ewokscore",
            "description": "Test 3",
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
        {
            "task_type": "class",
            "task_identifier": "ewokscore.tests.discover.module2.MyTask4",
            "required_input_names": ["a", "b"],
            "optional_input_names": ["c", "d"],
            "output_names": [],
            "category": "ewokscore",
            "description": None,
            "input_model": "ewokscore.tests.discover.module2.Task4Inputs",
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
        {
            "task_type": "class",
            "task_identifier": "ewokscore.tests.discover.module2.MyTask5",
            "required_input_names": [],
            "optional_input_names": [],
            "output_names": ["a", "b", "c", "d"],
            "category": "ewokscore",
            "description": None,
            "input_model": None,
            "output_model": "ewokscore.tests.discover.module2.Task5Outputs",
            "n_required_positional_inputs": 0,
        },
    ]

    METHOD_TASKS = [
        {
            "task_type": "method",
            "task_identifier": "ewokscore.tests.discover.module1.run",
            "required_input_names": ["a"],
            "optional_input_names": ["b"],
            "output_names": ["return_value"],
            "category": "ewokscore",
            "description": "Test 2",
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
        {
            "task_type": "method",
            "task_identifier": "ewokscore.tests.discover.module1.myfunc",
            "required_input_names": ["a"],
            "optional_input_names": ["b"],
            "output_names": ["return_value"],
            "category": "ewokscore",
            "description": None,
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
        {
            "task_type": "method",
            "task_identifier": "ewokscore.tests.discover.module1.func_with_pos",
            "required_input_names": ["c"],
            "optional_input_names": [],
            "output_names": ["return_value"],
            "category": "ewokscore",
            "description": None,
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 2,
        },
        {
            "task_type": "method",
            "task_identifier": "ewokscore.tests.discover.module2.run",
            "required_input_names": ["z", "c"],
            "optional_input_names": ["x", "d"],
            "output_names": ["return_value"],
            "category": "ewokscore",
            "description": "Test",
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
        {
            "task_type": "method",
            "task_identifier": "ewokscore.tests.discover.module2.myfunc",
            "required_input_names": ["z", "c"],
            "optional_input_names": ["x", "d"],
            "output_names": ["return_value"],
            "category": "ewokscore",
            "description": None,
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
    ]

    PPFMETHOD_TASKS = [
        {
            "task_type": "ppfmethod",
            "task_identifier": "ewokscore.tests.discover.module1.run",
            "required_input_names": ["a"],
            "optional_input_names": ["b"],
            "output_names": ["return_value"],
            "category": "ewokscore",
            "description": "Test 2",
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
        {
            "task_type": "ppfmethod",
            "task_identifier": "ewokscore.tests.discover.module2.run",
            "required_input_names": ["z", "c"],
            "optional_input_names": ["x", "d"],
            "output_names": ["return_value"],
            "category": "ewokscore",
            "description": "Test",
            "input_model": None,
            "output_model": None,
            "n_required_positional_inputs": 0,
        },
    ]

    TASKS: List[TaskDict] = [*CLASS_TASKS, *METHOD_TASKS, *PPFMETHOD_TASKS]

    if task_type is None and module is None:
        return TASKS

    if module is None:
        return [task for task in TASKS if task["task_type"] == task_type]

    if task_type is None:
        return [task for task in TASKS if module in task["task_identifier"]]

    return [
        task
        for task in TASKS
        if task["task_type"] == task_type and module in task["task_identifier"]
    ]


@pytest.fixture()
def engine():
    return CoreWorkflowEngine()



from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from .events import job_decorator as execute_graph_decorator
from .graph import TaskGraph
from .graph import load_graph as _load_graph
from .graph.execute import sequential
from .graph.graph_io import update_default_inputs
from .graph.serialize import GraphRepresentation
from .node import NodeIdType
from .task import Task

__all__ = [
    "execute_graph",
    "load_graph",
    "save_graph",
    "convert_graph",
    "graph_is_supported",
    "execute_graph_decorator",
]


def load_graph(
    graph: Any,
    inputs: Optional[List[dict]] = None,
    representation: Optional[Union[GraphRepresentation, str]] = None,
    root_dir: Optional[Union[str, Path]] = None,
    root_module: Optional[str] = None,
) -> TaskGraph:
    taskgraph = _load_graph(
        source=graph,
        representation=representation,
        root_dir=root_dir,
        root_module=root_module,
    )
    if inputs:
        update_default_inputs(taskgraph.graph, inputs)
    return taskgraph


def save_graph(
    graph: TaskGraph,
    destination,
    representation: Optional[Union[GraphRepresentation, str]] = None,
    **save_options,
) -> Union[str, dict]:
    return graph.dump(destination, representation=representation, **save_options)


def convert_graph(
    source,
    destination,
    inputs: Optional[List[dict]] = None,
    load_options: Optional[dict] = None,
    save_options: Optional[dict] = None,
) -> Union[str, dict]:
    if load_options is None:
        load_options = dict()
    if save_options is None:
        save_options = dict()
    graph = load_graph(source, inputs=inputs, **load_options)
    return save_graph(graph, destination, **save_options)


@execute_graph_decorator(engine="core")
def execute_graph(
    graph,
    inputs: Optional[List[dict]] = None,
    load_options: Optional[dict] = None,
    varinfo: Optional[dict] = None,
    execinfo: Optional[dict] = None,
    task_options: Optional[dict] = None,
    raise_on_error: Optional[bool] = True,
    outputs: Optional[List[dict]] = None,
    merge_outputs: Optional[bool] = True,
    output_tasks: Optional[bool] = False,
) -> Union[Dict[NodeIdType, Task], Dict[str, Any]]:
    if load_options is None:
        load_options = dict()
    taskgraph = load_graph(graph, inputs=inputs, **load_options)
    return sequential.execute_graph(
        taskgraph.graph,
        varinfo=varinfo,
        execinfo=execinfo,
        task_options=task_options,
        raise_on_error=raise_on_error,
        outputs=outputs,
        merge_outputs=merge_outputs,
        output_tasks=output_tasks,
    )


def graph_is_supported(graph: TaskGraph) -> bool:
    return not graph.is_cyclic and not graph.has_conditional_links



from abc import ABC
from abc import abstractmethod
from pathlib import Path
from typing import Any
from typing import List
from typing import Optional
from typing import Union

from .events.contexts import RawExecInfoType
from .graph import TaskGraph


class WorkflowEngine(ABC):
    """Python projects that provide Ewoks engines for deserializing, serializing and executing
    computational Ewoks graphs can implement this interface.

    To make it discoverable it can be added as an entry-point the the project. For
    example in a `pyproject.toml` file:

    .. code-block: toml

        [project.entry-points."ewoks.engines"]
        "<engine-name>" = "<project-name>.engine:MyWorkflowEngine"
    """

    @abstractmethod
    def execute_graph(
        self,
        graph: TaskGraph,
        *,
        inputs: Optional[List[dict]] = None,
        load_options: Optional[dict] = None,
        varinfo: Optional[dict] = None,
        execinfo: Optional[RawExecInfoType] = None,
        task_options: Optional[dict] = None,
        outputs: Optional[List[dict]] = None,
        merge_outputs: Optional[bool] = True,
        # Engine specific:
        **execute_options,
    ) -> Optional[dict]:
        """Execute a computional Ewoks graph."""
        pass


class WorkflowEngineWithSerialization(WorkflowEngine):
    """Ewoks engines with graph serialization capabilities."""

    @abstractmethod
    def deserialize_graph(
        self,
        graph: Any,
        *,
        inputs: Optional[List[dict]] = None,
        representation: Optional[str] = None,
        root_dir: Optional[Union[str, Path]] = None,
        root_module: Optional[str] = None,
        # Serializer specific:
        **deserialize_options,
    ) -> TaskGraph:
        """Convert a computational graph representation to the canonical in-memory representation `TaskGraph`."""
        pass

    @abstractmethod
    def serialize_graph(
        self,
        graph: TaskGraph,
        destination: Any,
        *,
        representation: Optional[str] = None,
        # Serializer specific:
        **serialize_options,
    ) -> Any:
        """Convert the canonical computational graph representation `TaskGraph` to another representation."""
        pass

    @abstractmethod
    def get_graph_representation(self, graph: Any) -> Optional[str]:
        """Return the representation if the engine recognizes the graph object."""
        pass



from copy import deepcopy
from typing import Union

from pydantic import Field

from ..bindings import execute_graph
from ..bindings import load_graph
from ..graph import inputs
from ..graph.taskgraph import TaskGraph
from ..missing_data import MISSING_DATA
from ..model import BaseInputModel
from ..task import Task


def test_shorten_task_identifiers():
    task_identifiers = ["a.b.c", "a.b.d", "a.b.e"]
    shortmap = inputs._shorten_task_identifiers(task_identifiers)
    expected = {"a.b.c": "c", "a.b.d": "d", "a.b.e": "e"}
    assert shortmap == expected

    task_identifiers = ["a.b.c", "a.b.d", "a.bb.c"]
    shortmap = inputs._shorten_task_identifiers(task_identifiers)
    expected = {"a.b.c": "b.c", "a.b.d": "d", "a.bb.c": "bb.c"}
    assert shortmap == expected

    task_identifiers = ["a.b.c", "a.b.c", "a.bb.c"]
    shortmap = inputs._shorten_task_identifiers(task_identifiers)
    expected = {"a.b.c": "b.c", "a.bb.c": "bb.c"}
    assert shortmap == expected


def test_graph_inputs():
    graph = create_graph()
    node_inputs = inputs.graph_inputs(graph)

    expected = [
        inputs.NodeInput(
            id="task1",
            label=None,
            task_identifier="ClassExample",
            name="a",
            value=1,
            required=True,
            description=None,
            examples=None,
            import_error=None,
        ),
        inputs.NodeInput(
            id="task1",
            label=None,
            task_identifier="ClassExample",
            name="b",
            value=MISSING_DATA,
            required=True,
            description=None,
            examples=None,
            import_error=None,
        ),
        inputs.NodeInput(
            id="task1",
            label=None,
            task_identifier="ClassExample",
            name="c",
            value=MISSING_DATA,
            required=False,
            description=None,
            examples=None,
            import_error=None,
        ),
        inputs.NodeInput(
            id="task1",
            label=None,
            task_identifier="ClassExample",
            name="d",
            value=MISSING_DATA,
            required=False,
            description=None,
            examples=None,
            import_error=None,
        ),
        inputs.NodeInput(
            id="task2",
            label="task2 label",
            task_identifier="ClassExampleWithModel",
            name="c",
            value=4,
            required=False,
            description="parameter c",
            examples=None,
            import_error=None,
        ),
        inputs.NodeInput(
            id="task2",
            label="task2 label",
            task_identifier="ClassExampleWithModel",
            name="d",
            value=-2,
            required=False,
            description=None,
            examples=[100, "word"],
            import_error=None,
        ),
        inputs.NodeInput(
            id="task3",
            label=None,
            task_identifier="method_example",
            name="e",
            value=5,
            required=True,
            description=None,
            examples=None,
            import_error=None,
        ),
        inputs.NodeInput(
            id="task3",
            label=None,
            task_identifier="method_example",
            name="g",
            value=MISSING_DATA,
            required=True,
            description=None,
            examples=None,
            import_error=None,
        ),
        inputs.NodeInput(
            id="task3",
            label=None,
            task_identifier="method_example",
            name="f",
            value=-3,
            required=False,
            description=None,
            examples=None,
            import_error=None,
        ),
        inputs.NodeInput(
            id="task4",
            label=None,
            task_identifier="NonExistingClass",
            name="guess",
            value=999,
            required=True,
            description=None,
            examples=None,
            import_error=ModuleNotFoundError("No module named 'does'"),
        ),
    ]

    # Patch exceptions for comparison
    for node_input in node_inputs:
        if node_input.import_error:
            node_input.import_error = repr(node_input.import_error)
    for node_input in expected:
        if node_input.import_error:
            node_input.import_error = repr(node_input.import_error)

    assert node_inputs == expected


def test_graph_inputs_as_table():
    graph = create_graph()
    column_names, rows, metadata, footnotes = inputs.graph_inputs_as_table(graph)

    expected_column_names = [
        "Name",
        "Value",
        "Description",
        "Examples",
        "Task identifier",
        "Id",
        "Label",
    ]
    expected_rows = [
        ["b⁽*⁾", "<MISSING_DATA>", "", "", "ClassExample", "task1", ""],
        ["g⁽*⁾", "<MISSING_DATA>", "", "", "method_example", "task3", ""],
        ["a", "1", "", "", "ClassExample", "task1", ""],
        ["c", "<MISSING_DATA>", "", "", "ClassExample", "task1", ""],
        ["d", "<MISSING_DATA>", "", "", "ClassExample", "task1", ""],
        ["c", "4", "parameter c", "", "ClassExampleWithModel", "task2", "task2 label"],
        [
            "d",
            "-2",
            "",
            "• 100\n• 'word'",
            "ClassExampleWithModel",
            "task2",
            "task2 label",
        ],
        ["e", "5", "", "", "method_example", "task3", ""],
        ["f", "-3", "", "", "method_example", "task3", ""],
        ["guess⁽†⁾", "999", "", "", "NonExistingClass", "task4", ""],
    ]
    expected_metadata = {"id": "testgraph", "description": "Test graph inputs"}
    expected_footnotes = [
        "⁽*⁾ Value is required for execution.",
        "⁽†⁾ Information from workflow only (task cannot be imported).",
    ]

    assert rows == expected_rows
    assert column_names == expected_column_names
    assert metadata == expected_metadata
    assert footnotes == expected_footnotes


class ClassExample(
    Task,
    input_names=["a", "b"],
    optional_input_names=["c", "d"],
    output_names=["a", "b", "c", "d"],
):
    def run(self):
        self.outputs.a = self.inputs.a
        self.outputs.b = self.inputs.b
        self.outputs.c = self.get_input_value("c", -1)
        self.outputs.d = self.get_input_value("d", -2)


class InputModel(BaseInputModel):
    a: Union[int, str] = Field(...)
    a: Union[int, str] = Field(...)
    c: Union[int, str] = Field(-1, description="parameter c")
    d: Union[int, str] = Field(-2, examples=[100, "word"])


class ClassExampleWithModel(
    Task,
    input_model=InputModel,
    output_names=["a", "b", "c", "d"],
):
    def run(self):
        self.outputs.a = self.inputs.a
        self.outputs.b = self.inputs.b
        self.outputs.c = self.inputs.c
        self.outputs.d = self.inputs.d


def method_example(a, b, e, g, c=-1, d=-2, f=-3):
    return {"a": a, "b": b, "c": c, "d": d, "e": e, "f": f, "g": g}


def create_graph() -> TaskGraph:
    graph = {"id": "testgraph", "schema_version": "1.1", "label": "Test graph inputs"}
    nodes = [
        {
            "id": "task1",
            "default_inputs": [{"name": "a", "value": 1}],
            "task_type": "class",
            "task_identifier": f"{__name__}.ClassExample",
        },
        {
            "id": "task2",
            "label": "task2 label",
            "default_inputs": [{"name": "b", "value": 3}, {"name": "c", "value": 4}],
            "task_type": "class",
            "task_identifier": f"{__name__}.ClassExampleWithModel",
        },
        {
            "id": "task3",
            "default_inputs": [{"name": "e", "value": 5}],
            "task_type": "method",
            "task_identifier": f"{__name__}.method_example",
        },
    ]

    links = [
        {
            "source": "task1",
            "target": "task2",
            "data_mapping": [{"source_output": "b", "target_input": "a"}],
        },
        {
            "source": "task2",
            "target": "task3",
            "map_all_data": True,
        },
    ]

    graph_dict = {"graph": graph, "links": links, "nodes": nodes}
    result = execute_graph(
        deepcopy(graph_dict),
        merge_outputs=False,
        inputs=[
            {"id": "task1", "name": "b", "value": 2},
            {"id": "task3", "name": "e", "value": 5},
            {"id": "task3", "name": "g", "value": 6},
        ],
        outputs=[{"all": True}],
    )
    expected = {
        "task1": {"a": 1, "b": 2, "c": -1, "d": -2},
        "task2": {"a": 2, "b": 3, "c": 4, "d": -2},
        "task3": {
            "return_value": {"a": 2, "b": 3, "c": 4, "d": -2, "e": 5, "g": 6, "f": -3}
        },
    }
    assert result == expected

    nodes.append(
        {
            "id": "task4",
            "default_inputs": [{"name": "guess", "value": 999}],
            "task_type": "class",
            "task_identifier": "does.not.exists.NonExistingClass",
        }
    )

    task_graph = load_graph(graph_dict)
    return task_graph



import cProfile
import os
import random
import re
import time
from contextlib import ExitStack
from contextlib import contextmanager
from typing import Any
from typing import Generator
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Type
from typing import Union

from ewoksutils.deprecation_utils import deprecated

from . import events
from . import missing_data
from . import node
from .hashing import UniversalHashable
from .model import BaseInputModel
from .model import BaseOutputModel
from .registration import Registered
from .variable import ReadOnlyVariableContainerNamespace
from .variable import VariableContainer
from .variable import VariableContainerMissingNamespace
from .variable import VariableContainerNamespace


class TaskInputError(ValueError):
    pass


class Task(Registered, UniversalHashable, register=False):
    """Node in a task Graph with named inputs and outputs.

    The universal hash of the task is equal to the universal
    hash of the output. The universal hash of the output is
    equal to the hash of the inputs and the task nonce.

    A task is done when its output exists.

    This is an abstract class. Instantiating a `Task` should be
    done with `ewokscore.inittask.instantiate_task`.
    """

    _INPUT_NAMES: Set[str] = set()
    _OPTIONAL_INPUT_NAMES: Set[str] = set()
    _OUTPUT_NAMES: Set[str] = set()
    _N_REQUIRED_POSITIONAL_INPUTS: int = 0
    _INPUT_MODEL: Optional[Type[BaseInputModel]] = None
    _OUTPUT_MODEL: Optional[Type[BaseOutputModel]] = None

    def __init__(
        self,
        inputs: Optional[Mapping] = None,
        varinfo: Optional[dict] = None,
        node_id: Optional[node.NodeIdType] = None,
        node_attrs: Optional[dict] = None,
        execinfo: Optional[dict] = None,
        profile_directory: Optional[dict] = None,
    ):
        """The named arguments are inputs and Variable configuration"""
        if inputs is None:
            inputs = dict()
        elif not isinstance(inputs, Mapping):
            raise TypeError(inputs, type(inputs))

        # Required outputs for the task to be "done"
        ovars = {varname: self.MISSING_DATA for varname in self._OUTPUT_NAMES}

        # Node/task info
        node_id = node.get_node_id(node_id, node_attrs)
        self.__node_id = node_id
        self.__node_label = node.get_node_label(node_id, node_attrs)
        self.__task_identifier = self._get_task_identifier(inputs)
        task_id = self.class_registry_name()
        task_id = node.get_task_identifier(node_attrs, task_id)
        self.__task_id = task_id
        if node_id and task_id:
            self.__execinfo = execinfo
        else:
            self.__execinfo = None

        # Misc
        self.__exception = None
        self.__succeeded = None
        self._cancelled = False
        self._profile_directory = profile_directory or dict()

        # The output hash will update dynamically if any of the input
        # variables change
        varinfo = node.get_varinfo(node_attrs, varinfo)
        inputs = self._check_inputs(inputs)

        self.__inputs = VariableContainer(value=inputs, varinfo=varinfo)
        self.__inputs_namespace = ReadOnlyVariableContainerNamespace(self.__inputs)
        self.__missing_inputs_namespace = VariableContainerMissingNamespace(
            self.__inputs
        )

        self.__outputs = VariableContainer(
            value=ovars,
            pre_uhash=self.__inputs,
            instance_nonce=self.class_nonce(),
            varinfo=varinfo,
        )
        self.__outputs_namespace = VariableContainerNamespace(self.__outputs)
        self.__missing_outputs_namespace = VariableContainerMissingNamespace(
            self.__outputs
        )

        # The task class has the same hash as its output
        super().__init__(pre_uhash=self.__outputs)

    def _check_inputs(self, inputs: Mapping) -> Mapping:
        """Check inputs without accessing the input values.
        Persisted variables are not loaded.
        """
        input_names = set(inputs.keys())

        # Check required inputs
        missing_required = self.required_input_names() - input_names
        if missing_required:
            self._raise_task_input_error("Missing inputs", str(list(missing_required)))

        # Check required positional inputs
        nrequiredargs = self._N_REQUIRED_POSITIONAL_INPUTS
        for i in range(nrequiredargs):
            if i not in inputs and str(i) not in inputs:
                self._raise_task_input_error(
                    "Missing inputs", f"positional argument #{i}"
                )

        # Init missing optional inputs
        missing_optional = self.optional_input_names() - input_names
        if missing_optional:
            inputs = dict(inputs)
            for varname in missing_optional:
                inputs[varname] = self.MISSING_DATA

        return inputs

    def _validate_inputs(self) -> None:
        """Check inputs with accessing the input values.
        Persisted variables are loaded.

        :raises pydantic.ValidationError:
        """
        if self._INPUT_MODEL is None:
            raise ValueError(
                "Trying to validate inputs while no input model was specified"
            )
        inputs = self.__inputs.get_variable_values()
        model = self._INPUT_MODEL(**inputs)

        for name in self._INPUT_MODEL.model_fields.keys():
            self.__inputs[name].value = getattr(model, name)

    def _validate_outputs(self) -> None:
        """Check outputs with accessing the output values.
        Persisted variables are loaded.

        :raises pydantic.ValidationError:
        """
        if self._OUTPUT_MODEL is None:
            raise ValueError(
                "Trying to validate outputs while no output model was specified"
            )
        outputs = self.__outputs.get_variable_values()
        model = self._OUTPUT_MODEL(**outputs)

        for name in self._OUTPUT_MODEL.model_fields.keys():
            self.__outputs[name].value = getattr(model, name)

    def __init_subclass__(
        subclass,
        input_names: Sequence[str] = tuple(),
        optional_input_names: Sequence[str] = tuple(),
        output_names: Sequence[str] = tuple(),
        n_required_positional_inputs: int = 0,
        input_model: Union[Type[BaseInputModel], None] = None,
        output_model: Union[Type[BaseInputModel], None] = None,
        **kwargs,
    ):
        super().__init_subclass__(**kwargs)

        input_names_set, optional_input_names_set = subclass._generate_inputs_sets(
            input_names,
            optional_input_names,
            n_required_positional_inputs,
            input_model,
        )
        output_names_set = subclass._generate_outputs_set(output_names, output_model)

        reserved = subclass._reserved_variable_names()
        forbidden = input_names_set & reserved
        forbidden |= optional_input_names_set & reserved
        forbidden |= output_names_set & reserved
        if forbidden:
            raise RuntimeError(
                "The following names cannot be used a variable names: "
                + str(list(forbidden))
            )

        # Ensures that each subclass has their own sets:
        subclass._INPUT_NAMES = subclass._INPUT_NAMES | input_names_set
        subclass._OPTIONAL_INPUT_NAMES = (
            subclass._OPTIONAL_INPUT_NAMES | optional_input_names_set
        )
        subclass._OUTPUT_NAMES = subclass._OUTPUT_NAMES | output_names_set
        subclass._N_REQUIRED_POSITIONAL_INPUTS = n_required_positional_inputs
        subclass._INPUT_MODEL = input_model
        subclass._OUTPUT_MODEL = output_model

    @classmethod
    def _generate_inputs_sets(
        subclass,
        input_names: Sequence[str],
        optional_input_names: Sequence[str],
        n_required_positional_inputs: int,
        input_model: Union[Type[BaseInputModel], None],
    ) -> Tuple[Set[str], Set[str]]:
        if input_model is None:
            input_names_set = set(input_names)
            optional_input_names_set = set(optional_input_names)

            has_input_names = bool(
                input_names_set
                or optional_input_names_set
                or n_required_positional_inputs > 0
            )
            if has_input_names and subclass._INPUT_MODEL is not None:
                raise TypeError(
                    f"""Cannot use input_names or optional_input_names since the original task {subclass} uses a input model.
                    Specify inputs via a subclass of the original task input model."""
                )

            return input_names_set, optional_input_names_set

        if not issubclass(input_model, BaseInputModel):
            raise TypeError(
                "input_model should be a subclass of ewokscore.model.BaseInputModel"
            )

        if input_names or optional_input_names or n_required_positional_inputs:
            raise TypeError(
                "input_model cannot be used with input_names, optional_input_names or n_required_positional_inputs. Please use one or the other"
            )

        subclass_has_input_names = bool(
            subclass._INPUT_NAMES
            or subclass._OPTIONAL_INPUT_NAMES
            or subclass._N_REQUIRED_POSITIONAL_INPUTS > 0
        )
        if subclass_has_input_names and subclass._INPUT_MODEL is None:
            raise TypeError(
                f"""Cannot use input_model since the original task {subclass} uses input_names and/or n_required_positional_inputs.
                Specify inputs via a input_names or optional_input_names."""
            )

        if subclass._INPUT_MODEL is not None and not issubclass(
            input_model, subclass._INPUT_MODEL
        ):
            raise TypeError(
                f"Input model {input_model} from task subclass must be a subclass of the original task input model {subclass._INPUT_MODEL}"
            )

        fields = input_model.model_fields
        return (
            set(name for name, field in fields.items() if field.is_required()),
            set(name for name, field in fields.items() if not field.is_required()),
        )

    @classmethod
    def _generate_outputs_set(
        subclass,
        output_names: Sequence[str],
        output_model: Union[Type[BaseOutputModel], None],
    ) -> Set[str]:
        if output_model is None:
            output_names_set = set(output_names)

            has_output_names = bool(output_names_set)
            if has_output_names and subclass._OUTPUT_MODEL is not None:
                raise TypeError(
                    f"""Cannot use output_names since the original task {subclass} uses a output model.
                    Specify outputs via a subclass of the original task output model."""
                )

            return output_names_set

        if not issubclass(output_model, BaseOutputModel):
            raise TypeError(
                "output_model should be a subclass of ewokscore.model.BaseOutputModel"
            )

        if output_names:
            raise TypeError(
                "output_model cannot be used with output_names. Please use one or the other"
            )

        subclass_has_output_names = bool(subclass._OUTPUT_NAMES)
        if subclass_has_output_names and subclass._OUTPUT_MODEL is None:
            raise TypeError(
                f"""Cannot use output_model since the original task {subclass} uses output_names.
                Specify outputs via a output_names."""
            )

        if subclass._OUTPUT_MODEL is not None and not issubclass(
            output_model, subclass._OUTPUT_MODEL
        ):
            raise TypeError(
                f"Output model {output_model} from task subclass must be a subclass of the original task output model {subclass._OUTPUT_MODEL}"
            )

        fields = output_model.model_fields
        return set(fields)

    @staticmethod
    def _reserved_variable_names():
        return VariableContainerNamespace._reserved_variable_names()

    @classmethod
    def instantiate(cls, registry_name: str, **kw):
        r"""Factory method for instantiating a derived class.

        :param str registry_name: for example "tasklib.tasks.MyTask" or "MyTask"
        :param \**kw: `Task` constructor arguments
        :returns Task:
        """
        return cls.get_subclass(registry_name)(**kw)

    @classmethod
    def required_input_names(cls) -> Set[str]:
        if cls._INPUT_MODEL:
            return {
                name
                for name, field in cls._INPUT_MODEL.model_fields.items()
                if field.is_required()
            }
        return cls._INPUT_NAMES

    @classmethod
    def optional_input_names(cls) -> Set[str]:
        if cls._INPUT_MODEL:
            return {
                name
                for name, field in cls._INPUT_MODEL.model_fields.items()
                if not field.is_required()
            }
        return cls._OPTIONAL_INPUT_NAMES

    @classmethod
    def input_names(cls) -> Set[str]:
        if cls._INPUT_MODEL:
            return set(cls._INPUT_MODEL.model_fields)
        return cls._INPUT_NAMES | cls._OPTIONAL_INPUT_NAMES

    @classmethod
    def output_names(cls) -> Set[str]:
        return cls._OUTPUT_NAMES

    @classmethod
    def input_model(cls) -> Optional[Type[BaseInputModel]]:
        return cls._INPUT_MODEL

    @classmethod
    def output_model(cls) -> Optional[Type[BaseOutputModel]]:
        return cls._OUTPUT_MODEL

    @classmethod
    def n_required_positional_inputs(cls) -> int:
        return cls._N_REQUIRED_POSITIONAL_INPUTS

    @classmethod
    def class_nonce_data(cls):
        return super().class_nonce_data() + (
            sorted(cls.input_names()),
            sorted(cls.output_names()),
            cls._N_REQUIRED_POSITIONAL_INPUTS,
        )

    @property
    def input_variables(self) -> VariableContainer:
        if self.__inputs is None:
            raise RuntimeError("references have been removed")
        return self.__inputs

    @property
    def inputs(self) -> ReadOnlyVariableContainerNamespace:
        return self.__inputs_namespace

    @property
    def missing_inputs(self) -> VariableContainerMissingNamespace:
        return self.__missing_inputs_namespace

    def get_input_value(self, key, default: Any = missing_data.MISSING_DATA) -> Any:
        if self.missing_inputs[key]:
            return default
        return self.inputs[key]

    @property
    def input_uhashes(self):
        return self.get_input_uhashes()

    def get_input_uhashes(self):
        return self.__inputs.get_variable_uhashes()

    @property
    @deprecated(
        "the property 'input_values' is deprecated in favor of the function 'get_input_values'"
    )
    def input_values(self):
        """DEPRECATED"""
        return self.get_input_values()

    def get_input_values(self):
        return self.__inputs.get_variable_values()

    @property
    @deprecated(
        "the property 'named_input_values' is deprecated in favor of the function 'get_named_input_values'"
    )
    def named_input_values(self):
        """DEPRECATED"""
        return self.get_named_input_values()

    def get_named_input_values(self):
        return self.__inputs.get_named_variable_values()

    @property
    @deprecated(
        "the property 'positional_input_values' is deprecated in favor of the function 'get_positional_input_values'"
    )
    def positional_input_values(self):
        """DEPRECATED"""
        return self.__inputs.get_positional_input_values()

    def get_positional_input_values(self):
        return self.__inputs.get_positional_variable_values()

    @property
    @deprecated(
        "the property 'npositional_inputs' is deprecated in favor of the property 'n_positional_inputs'"
    )
    def npositional_inputs(self):
        """DEPRECATED"""
        return self.n_positional_inputs

    @property
    def n_positional_inputs(self) -> int:
        return self.__inputs.n_positional_variables

    @property
    def output_variables(self) -> VariableContainer:
        return self.__outputs

    @property
    def missing_outputs(self) -> VariableContainerMissingNamespace:
        return self.__missing_outputs_namespace

    @property
    def outputs(self) -> VariableContainerNamespace:
        return self.__outputs_namespace

    def get_output_value(self, key, default: Any = missing_data.MISSING_DATA) -> Any:
        if self.missing_outputs[key]:
            return default
        return self.outputs[key]

    @property
    @deprecated(
        "the property 'output_uhashes' is deprecated in favor of the function 'get_output_uhashes'"
    )
    def output_uhashes(self):
        """DEPRECATED"""
        return self.get_output_uhashes()

    def get_output_uhashes(self):
        return self.__outputs.get_variable_uhashes()

    @property
    @deprecated(
        "the property 'output_values' is deprecated in favor of the function 'get_output_values'"
    )
    def output_values(self):
        """DEPRECATED"""
        return self.get_output_values()

    def get_output_values(self):
        return self.__outputs.get_variable_values()

    @property
    @deprecated(
        "the property 'output_transfer_data' is deprecated in favor of the function 'get_output_transfer_data'"
    )
    def output_transfer_data(self):
        """DEPRECATED"""
        return self.get_output_transfer_data()

    def get_output_transfer_data(self):
        """The values are either `DataUri` or `Variable`"""
        return self.__outputs.get_variable_transfer_data()

    @property
    def output_metadata(self) -> Union[dict, None]:
        return self.__outputs.metadata

    def _update_output_metadata(self):
        metadata = self.output_metadata
        if metadata is None:
            return
        if self.__node_label:
            metadata.setdefault("title", self.__node_label)

    @property
    def done(self) -> bool:
        """Completed (with or without exception)"""
        return self.failed or self.succeeded

    @property
    def succeeded(self) -> bool:
        """Completed without exception and with output values"""
        if self._OUTPUT_NAMES:
            return self.__outputs.has_value
        else:
            return self.__succeeded

    @property
    def failed(self) -> bool:
        """Completed with exception"""
        return self.__exception is not None

    @property
    def exception(self) -> Optional[Exception]:
        return self.__exception

    def _get_repr_data(self):
        data = super()._get_repr_data()
        if self.__node_label:
            data["label"] = repr(str(self.__node_label))
        else:
            data["label"] = None
        return data

    @property
    def label(self) -> str:
        if self.__node_label:
            return self.__node_label
        else:
            return str(self)

    @property
    def node_id(self) -> node.NodeIdType:
        return self.__node_id

    @property
    def task_identifier(self) -> str:
        return self.__task_identifier

    def _get_task_identifier(self, inputs: Mapping) -> str:
        return self.class_registry_name()

    @property
    def job_id(self) -> Optional[str]:
        if self.__execinfo:
            return self.__execinfo.get("job_id")

    @property
    def workflow_id(self) -> Optional[str]:
        if self.__execinfo:
            return self.__execinfo.get("workflow_id")

    @property
    def _profile_filename(self) -> Optional[str]:
        profile_directory = self._profile_directory
        if not profile_directory:
            return
        job_id = self.job_id
        workflow_id = self.workflow_id
        node_id = self.node_id
        if job_id is None or workflow_id is None or node_id is None:
            return
        if isinstance(node_id, tuple):
            node_id = "_".join(map(str, tuple))
        else:
            node_id = str(node_id)

        job_id = re.sub(r"[^A-Za-z0-9]", "_", job_id)
        workflow_id = re.sub(r"[^A-Za-z0-9]", "_", workflow_id)
        node_id = re.sub(r"[^A-Za-z0-9]", "_", node_id)

        timestamp = int(time.time() * 1000)
        random_chars = "".join(
            random.choices("abcdefghijklmnopqrstuvwxyz0123456789", k=8)
        )
        filename = f"{timestamp}_{random_chars}_{node_id}.prof"

        return os.path.join(profile_directory, workflow_id, job_id, filename)

    def _iter_missing_input_values(self):
        for iname in self._INPUT_NAMES:
            var = self.__inputs.get(iname)
            if var is None or not var.has_value:
                yield iname

    @property
    def is_ready_to_execute(self):
        try:
            next(iter(self._iter_missing_input_values()))
        except StopIteration:
            return True
        return False

    @property
    def cancelled(self) -> bool:
        """Return True if the task has been cancelled by the user"""
        return self._cancelled

    @cancelled.setter
    def cancelled(self, cancelled: bool) -> None:
        self._cancelled = cancelled

    def assert_ready_to_execute(self):
        lst = list(self._iter_missing_input_values())
        if lst:
            self._raise_task_input_error(
                "The following inputs could not be loaded", str(lst)
            )

    def _raise_task_input_error(self, prefix: str, message: str) -> None:
        self._raise_task_error(prefix, message, TaskInputError)

    def _raise_task_error(
        self,
        prefix: str,
        message: str,
        exc_class: Type[Exception],
        cause: Optional[Type[Exception]] = None,
    ) -> None:
        node_id = self.__node_id
        task_identifier = self.task_identifier
        if self.__node_label:
            err_msg = f"{prefix} for ewoks task {self.__node_label!r} (id: {node_id!r}, task: {task_identifier!r}): {message}"
        else:
            err_msg = f"{prefix} for ewoks task (id: {node_id!r}, task: {task_identifier!r}): {message}"
        raise exc_class(err_msg) from cause

    def reset_state(self):
        self._cancelled = False
        self.__exception = None
        self.__succeeded = None
        self.__outputs.reset()

    def execute(
        self,
        force_rerun: Optional[bool] = False,
        raise_on_error: Optional[bool] = True,
        cleanup_references: Optional[bool] = False,
    ):
        with ExitStack() as stack:
            ctx = self._profile_time()
            _ = stack.enter_context(ctx)

            ctx = events.node_context(
                self.__execinfo, node_id=self.__node_id, task_id=self.__task_id
            )
            self.__execinfo = stack.enter_context(ctx)

            self.reset_state()

            ctx = self._send_task_events()
            _ = stack.enter_context(ctx)

            try:
                if force_rerun:
                    # Rerun a task which is already done
                    self.__outputs.force_non_existing()
                if self.done:
                    return
                self.assert_ready_to_execute()

                if self._INPUT_MODEL:
                    self._validate_inputs()

                self.run()

                if self._OUTPUT_MODEL:
                    self._validate_outputs()
                self._update_output_metadata()
                self.__outputs.dump()
                self.__succeeded = True
            except Exception as e:
                self.__exception = e
                if raise_on_error:
                    self._raise_task_error(
                        "Execution failed", str(e), RuntimeError, cause=e
                    )
            finally:
                if cleanup_references:
                    self.cleanup_references()

    @contextmanager
    def _profile_time(self) -> Generator[None, None, None]:
        """Optional time profiling within this context."""
        _profile_filename = self._profile_filename
        if _profile_filename:
            profiler = cProfile.Profile()
            profiler.enable()
        try:
            yield
        finally:
            if _profile_filename:
                profiler.disable()
                os.makedirs(os.path.dirname(_profile_filename), exist_ok=True)
                profiler.dump_stats(_profile_filename)

    @contextmanager
    def _send_task_events(self) -> Generator[None, None, None]:
        """Send an ewoks start event on enter and stop event on exit."""
        self._send_start_event()
        try:
            yield
        finally:
            self._send_send_event()

    def _send_event(self, **kwargs):
        """Send an ewoks event"""
        if self.__execinfo:
            events.send_task_event(execinfo=self.__execinfo, **kwargs)

    def _send_start_event(self):
        input_uris = [
            {"name": name, "value": str(uri) if uri else None}
            for name, uri in self.__inputs.get_variable_uris().items()
        ]
        output_uris = [
            {"name": name, "value": str(uri) if uri else None}
            for name, uri in self.__outputs.get_variable_uris().items()
        ]
        task_uri = self.__outputs.data_uri
        if task_uri:
            task_uri = str(task_uri)
        self._send_event(
            event="start",
            input_uris=input_uris,
            output_uris=output_uris,
            task_uri=task_uri,
        )

    def _send_send_event(self):
        self._send_event(event="end", exception=self.exception)

    def cleanup_references(self):
        """Removes all references to the inputs.
        Side effect: fixes the uhash of the task and outputs
        """
        self.__inputs = None
        self.__inputs_namespace = None
        self.__missing_inputs_namespace = None
        self.__outputs.cleanup_references()
        super().cleanup_references()

    def run(self):
        """To be implemented by the derived classes"""
        raise NotImplementedError

    def cancel(self):
        """
        Function called when a task is cancelled.
        To be implemented by the derived classes
        """
        raise NotImplementedError



import inspect
import logging
import pkgutil
import sys
from fnmatch import fnmatch
from types import FunctionType
from types import ModuleType
from typing import Generator
from typing import List
from typing import Optional
from typing import TypedDict

from ewoksutils.import_utils import import_module
from ewoksutils.import_utils import qualname

from .entry_points import entry_points
from .task import Task


class _TaskInputs(TypedDict):
    required_input_names: List[str]
    optional_input_names: List[str]
    n_required_positional_inputs: int


class _CommonTaskFields(_TaskInputs):
    task_identifier: str
    output_names: List[str]
    category: str
    description: Optional[str]
    input_model: Optional[str]
    output_model: Optional[str]


class TaskDict(_CommonTaskFields):
    task_type: str


logger = logging.getLogger(__name__)


def discover_tasks_from_modules(
    *module_names_or_patterns: str,
    task_type: Optional[str] = None,
    reload: bool = False,
    raise_import_failure: bool = True,
) -> List[TaskDict]:
    if task_type is None:
        task_types = ("class", "ppfmethod", "method")
    else:
        task_types = (task_type,)

    result = list()
    for task_type in task_types:
        # Module names can contain patterns
        for module_name_or_pattern in module_names_or_patterns:
            for module_name in _iter_modules_from_pattern(
                module_name_or_pattern,
                reload=reload,
                raise_import_failure=raise_import_failure,
            ):

                result.extend(
                    _iter_discover_tasks_from_modules(
                        module_name,
                        task_type=task_type,
                        reload=reload,
                        raise_import_failure=raise_import_failure,
                    )
                )

    return result


def _iter_discover_tasks_from_modules(
    *module_names: str,
    task_type: str,
    reload: bool = False,
    raise_import_failure: bool = True,
) -> Generator[TaskDict, None, None]:
    if "" not in sys.path:
        # This happens when the python process was launched
        # through a python console script
        sys.path.append("")

    if task_type == "method":
        yield from _iter_method_tasks(
            *module_names, reload=reload, raise_import_failure=raise_import_failure
        )
    elif task_type == "ppfmethod":
        yield from _iter_ppfmethod_tasks(
            *module_names, reload=reload, raise_import_failure=raise_import_failure
        )
    elif task_type == "class":
        for module_name in module_names:
            _safe_import_module(
                module_name, reload=reload, raise_import_failure=raise_import_failure
            )
        yield from _iter_registered_tasks(*module_names)
    else:
        raise ValueError(f"Task type {task_type} does not support discovery")


def _iter_registered_tasks(*filter_modules: str) -> Generator[TaskDict, None, None]:
    """Yields all task classes registered in the current process."""
    for cls in Task.get_subclasses():
        assert issubclass(cls, Task)
        module = cls.__module__
        if filter_modules and not any(
            module.startswith(prefix) for prefix in filter_modules
        ):
            continue

        task_identifier = cls.class_registry_name()
        if task_identifier is None:
            # Exclude unregistered tasks
            continue

        category = task_identifier.split(".")[0]
        name = task_identifier.split(".")[-1]
        if name.startswith("_"):
            # Exclude hidden tasks
            continue

        input_model = cls.input_model()
        output_model = cls.output_model()
        yield {
            "task_type": "class",
            "task_identifier": task_identifier,
            "required_input_names": sorted(cls.required_input_names()),
            "optional_input_names": sorted(cls.optional_input_names()),
            "output_names": sorted(cls.output_names()),
            "category": category,
            "description": cls.__doc__,
            "input_model": qualname(input_model) if input_model else None,
            "output_model": qualname(output_model) if output_model else None,
            "n_required_positional_inputs": cls.n_required_positional_inputs(),
        }


def _iter_method_tasks(
    *module_names: str,
    reload: bool = False,
    raise_import_failure: bool = False,
) -> Generator[TaskDict, None, None]:
    """Yields all task methods from the provided module_names. The module_names will be will
    imported for discovery.
    """
    for module_name in module_names:
        mod = _safe_import_module(
            module_name, reload=reload, raise_import_failure=raise_import_failure
        )
        if mod is None:
            continue
        for method_name, method_qn in inspect.getmembers(mod, inspect.isfunction):
            if method_name.startswith("_"):
                continue

            yield {
                "task_type": "method",
                **_common_method_task_fields(method_name, method_qn, mod),
            }


def _iter_ppfmethod_tasks(
    *module_names: str,
    reload: bool = False,
    raise_import_failure: bool = False,
) -> Generator[TaskDict, None, None]:
    """Yields all task ppfmethods from the provided module_names. The module_names will be will
    imported for discovery.

    The difference with regular methods is that ppfmethods are expected to be called `run`. Other method names will be ignored.
    """
    for module_name in module_names:
        mod = _safe_import_module(
            module_name, reload=reload, raise_import_failure=raise_import_failure
        )
        if mod is None:
            continue
        for method_name, method_qn in inspect.getmembers(mod, inspect.isfunction):
            if method_name != "run":
                continue

            yield {
                "task_type": "ppfmethod",
                **_common_method_task_fields(method_name, method_qn, mod),
            }


def _iter_discover_all_tasks(
    reload: bool = False,
    task_type: Optional[str] = None,
    raise_import_failure: bool = False,
) -> Generator[TaskDict, None, None]:
    visited = set()
    if task_type is None:
        task_types = ("class", "ppfmethod", "method")
    else:
        task_types = (task_type,)

    for task_type in task_types:
        group = "ewoks.tasks." + task_type
        for entrypoint in entry_points(group):
            module_pattern = entrypoint.name
            if module_pattern is visited:
                continue
            visited.add(module_pattern)
            yield from discover_tasks_from_modules(
                module_pattern,
                task_type=task_type,
                reload=reload,
                raise_import_failure=raise_import_failure,
            )


def discover_all_tasks(
    reload: bool = False,
    task_type: Optional[str] = None,
    raise_import_failure: bool = False,
) -> List[TaskDict]:
    return list(
        _iter_discover_all_tasks(
            reload=reload,
            task_type=task_type,
            raise_import_failure=raise_import_failure,
        )
    )


def _iter_modules_from_pattern(
    module_pattern: str, reload: bool = False, raise_import_failure: bool = False
) -> Generator[str, None, None]:
    if "*" not in module_pattern:
        yield module_pattern
        return
    ndots = module_pattern.count(".")
    parts = module_pattern.split(".")
    pkg = _safe_import_module(
        parts[0], reload=reload, raise_import_failure=raise_import_failure
    )
    if pkg is None:
        return
    if raise_import_failure:

        def onerror(module_name):
            raise

    else:
        onerror = _onerror
    for pkginfo in pkgutil.walk_packages(
        pkg.__path__, pkg.__name__ + ".", onerror=onerror
    ):
        if pkginfo.name.count(".") == ndots and fnmatch(pkginfo.name, module_pattern):
            yield pkginfo.name


def _safe_import_module(
    module_name: str, reload: bool = False, raise_import_failure: bool = False
) -> Optional[ModuleType]:
    try:
        return import_module(module_name, reload=reload)
    except Exception as e:
        if raise_import_failure:
            raise
        _onerror(module_name, exception=e)


def _onerror(module_name, exception: Optional[Exception] = None):
    if exception is None:
        exception = sys.exc_info()[1]
    logger.error(f"Module '{module_name}' cannot be imported: {exception}")


def _method_arguments(method) -> _TaskInputs:
    sig = inspect.signature(method)
    required_input_names: List[str] = list()
    optional_input_names: List[str] = list()
    n_required_positional_inputs = 0

    for name, param in sig.parameters.items():
        if param.kind == param.POSITIONAL_ONLY:
            n_required_positional_inputs += 1
            continue
        if param.kind == param.VAR_POSITIONAL:
            continue
        if param.kind == param.VAR_KEYWORD:
            continue

        required = param.default is inspect._empty
        if required:
            required_input_names.append(name)
        else:
            optional_input_names.append(name)

    return {
        "required_input_names": required_input_names,
        "optional_input_names": optional_input_names,
        "n_required_positional_inputs": n_required_positional_inputs,
    }


def _common_method_task_fields(
    method_name: str, method_qn: FunctionType, mod: ModuleType
) -> _CommonTaskFields:

    task_identifier = qualname(method_qn)
    method = getattr(mod, method_name)

    return {
        **_method_arguments(method),
        "task_identifier": qualname(method_qn),
        "output_names": ["return_value"],
        "category": task_identifier.split(".")[0],
        "description": method.__doc__,
        "input_model": None,
        "output_model": None,
    }



import logging
from typing import Dict
from typing import Union

import networkx
from packaging.version import Version
from packaging.version import parse as parse_version

from .metadata import SchemaMetadata
from .update import from_v1_0_to_v1_1
from .update import v0_update

_VERSIONS = None


def get_versions() -> Dict[Version, SchemaMetadata]:
    global _VERSIONS
    if _VERSIONS is not None:
        return _VERSIONS

    _VERSIONS = {
        parse_version("0.0"): SchemaMetadata(("0.0", "0.0.1"), v0_update),
        parse_version("1.0"): SchemaMetadata(("0.1.0-rc", None), from_v1_0_to_v1_1),
        parse_version("1.1"): SchemaMetadata(("0.1.0-rc", None), None),
    }
    return _VERSIONS


# Major version: increment when changing the existing schema
# Minor version: increment when adding features or deprecating the existing schema
LATEST_VERSION = list(get_versions().keys())[-1]

# The default version may be set to something else if we don't want the latest version to be the default
DEFAULT_VERSION = LATEST_VERSION


logger = logging.getLogger(__name__)


def normalize_schema_version(graph: Union[dict, networkx.Graph]):
    if isinstance(graph, dict):
        graph_metadata = graph["graph"]
    else:
        graph_metadata = graph.graph
    schema_version = graph_metadata.get("schema_version", None)
    if schema_version:
        pversion = parse_version(schema_version)
    else:
        logger.info(
            'Graph has no "schema_version": assume version "%s"', DEFAULT_VERSION
        )
        pversion = DEFAULT_VERSION
    if pversion != LATEST_VERSION:
        # This warning is given because an exception may occur before `update_graph_schema`
        # is called due to the different schema version.
        logger.warning(
            'Graph schema version "%s" is not equal to the latest version "%s"',
            pversion,
            LATEST_VERSION,
        )

    graph_metadata["schema_version"] = str(pversion)


def update_graph_schema(graph: networkx.DiGraph):
    """
    Updates the graph description to a higher schema version or raises an
    exception.
    If the schema version is known it will provide library version bounds
    in the exception message.
    """
    schema_version = parse_version(graph.graph["schema_version"])

    schema_metadata = get_versions().get(schema_version)
    update_method = schema_metadata.update_method if schema_metadata else None
    if not update_method:
        raise GraphSchemaError(schema_version)

    before = schema_version
    try:
        update_method(graph)
    except Exception:
        raise GraphSchemaError(schema_version)
    else:
        after = parse_version(graph.graph["schema_version"])
        if before == after:
            raise RuntimeError("graph conversion did not update the schema version")
        if before > after:
            raise RuntimeError("graph conversion did not increment the schema version")


class GraphSchemaError(ValueError):
    def __init__(self, schema_version: Version) -> None:
        version_metadata = get_versions().get(schema_version)

        if not version_metadata:
            return super().__init__(
                f'Graph schema version "{schema_version}" is either invalid or requires a newer library version: python3 -m pip install --upgrade ewokscore'
            )
        lbound, ubound = version_metadata.ewokscore_bounds

        if not ubound:
            return super().__init__(
                f'Graph schema version "{schema_version}" requires another library version: python3 -m pip install "ewokscore>={lbound}"'
            )

        return super().__init__(
            f'Graph schema version "{schema_version}" requires another library version: python3 -m pip install "ewokscore>={lbound},<{ubound}"`'
        )



"""Manage the EWOKS event logger which is a global object"""

import logging
import os
from contextlib import contextmanager
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

from ewoksutils.logging_utils.asyncwrapper import AsyncHandlerWrapper
from ewoksutils.logging_utils.cleanup import cleanup_handler
from ewoksutils.logging_utils.cleanup import cleanup_logger
from ewoksutils.logging_utils.cleanup import protect_logging_state

from .handlers import instantiate_handler
from .handlers import is_ewoks_event_handler

_app_logger = logging.getLogger(__name__)
EWOKS_EVENT_LOGGER_NAME = f"__{__name__}__"
ENABLE_EWOKS_EVENTS_BY_DEFAULT = True


def send(
    *args,
    handlers: Optional[List[Dict[str, Any]]] = None,
    asynchronous: Optional[bool] = None,
    **kw,
) -> None:
    """Log an EWOKS event with the EWOKS event handlers and the application handlers."""
    with _ewoks_event_logger(
        handlers=handlers, asynchronous=asynchronous, cleanup_on_different_handlers=True
    ) as logger:
        # Send to the EWOKS event handlers
        logger.info(*args, **kw)
        # Send to the application log handlers
        _app_logger.info(*args, **kw)


def add_handler(
    handler: logging.Handler,
    asynchronous: Optional[bool] = None,
) -> None:
    """Add a handler to the global EWOKS event logger."""
    with _ewoks_event_logger() as logger:
        if _has_handler_instance(logger, handler):
            return
        if asynchronous is None and is_ewoks_event_handler(handler):
            asynchronous = handler.BLOCKING
        if asynchronous:
            handler = AsyncHandlerWrapper(handler)
        logger.addHandler(handler)


def remove_handler(handler: logging.Handler) -> None:
    """Remove a handler from all loggers that receive EWOKS events."""
    with _ewoks_event_logger() as logger:
        for linstance, hinstance in _iter_handler_owners(logger, handler):
            linstance.removeHandler(hinstance)
            cleanup_handler(hinstance)


def cleanup():
    """Pending events will be dropped"""
    with protect_logging_state():
        _cleanup_ewoks_event_logger()


def _after_fork_in_child():
    cleanup()


if hasattr(os, "register_at_fork"):
    os.register_at_fork(after_in_child=_after_fork_in_child)


@contextmanager
def _ewoks_event_logger(
    handlers: Optional[List[Dict[str, Any]]] = None,
    asynchronous: Optional[bool] = None,
    cleanup_on_different_handlers: bool = False,
) -> Iterator[logging.Logger]:
    """Initialize and yield the EWOKS event logger"""
    # Issue with logging and forking:
    # https://pythonspeed.com/articles/python-multiprocessing/

    with protect_logging_state():
        if _ewoks_event_logger_requires_cleanup(
            handlers=handlers,
            cleanup_on_different_handlers=cleanup_on_different_handlers,
        ):
            _cleanup_ewoks_event_logger()
        if _ewoks_event_logger_requires_init():
            _init_ewoks_event_logger(handlers, asynchronous)
        yield logging.getLogger(EWOKS_EVENT_LOGGER_NAME)


def _cleanup_ewoks_event_logger():
    """Cleanup and delete the global EWOKS event logger"""
    cleanup_logger(EWOKS_EVENT_LOGGER_NAME)


def _ewoks_event_logger_requires_init() -> bool:
    logger = logging.getLogger(EWOKS_EVENT_LOGGER_NAME)
    return not hasattr(logger, "ewoks_pid")


def _ewoks_event_logger_requires_cleanup(
    handlers: Optional[List[Dict[str, Any]]] = None,
    cleanup_on_different_handlers: bool = False,
) -> bool:
    logger = logging.getLogger(EWOKS_EVENT_LOGGER_NAME)

    ewoks_pid = getattr(logger, "ewoks_pid", None)
    if ewoks_pid is not None and ewoks_pid != os.getpid():
        # Process forked
        return True

    if cleanup_on_different_handlers:
        ewoks_handlers = getattr(logger, "ewoks_handlers", None)
        if ewoks_handlers != handlers:
            return True

    return False


def _init_ewoks_event_logger(
    handlers: Optional[List[Dict[str, Any]]], asynchronous: Optional[bool]
):
    logger = logging.getLogger(EWOKS_EVENT_LOGGER_NAME)
    logger.setLevel(logging.DEBUG)
    logger.ewoks_pid = os.getpid()
    logger.ewoks_handlers = handlers
    logger.propagate = False
    if not handlers:
        return
    for desc in handlers:
        try:
            kwargs = {
                arg["name"]: arg["value"] for arg in desc.get("arguments", list())
            }
            handler = instantiate_handler(desc["class"], **kwargs)
        except Exception as e:
            raise RuntimeError(
                "cannot create an EWOKS event handler from the description"
            ) from e
        asynchronous_handler = desc.get("asynchronous", asynchronous)
        add_handler(handler, asynchronous=asynchronous_handler)


def _iter_loggers(logger: logging.Logger) -> Iterator[logging.Logger]:
    """Yield all loggers which will receive EWOKS events."""
    _logger = logger
    while _logger is not None:
        yield _logger
        if not _logger.propagate:
            return
        _logger = _logger.parent


def _iter_handler_owners(
    logger: logging.Logger, instance: logging.Handler
) -> Iterator[Tuple[logging.Logger, logging.Handler]]:
    """Yield all loggers which have a specific handler (or a handler that wraps the specific event handler)."""
    for _logger in _iter_loggers(logger):
        for handler in _logger.handlers:
            if handler is instance:
                yield _logger, instance
            elif isinstance(handler, AsyncHandlerWrapper):
                instance2 = handler.wrapped_handler
                if instance2 is instance:
                    yield _logger, instance2


def _has_handler_instance(logger: logging.Logger, instance: logging.Handler) -> bool:
    """Is this handler registered with a logger or it's parents?"""
    for _ in _iter_handler_owners(logger, instance):
        return True
    return False



"""Define and parse ewoks events"""

import traceback
from datetime import datetime
from numbers import Number
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional

from ewoksutils.event_utils import FIELD_TYPES

from . import global_state
from .initialize_events import ExecInfoType


def send_job_event(**kw):
    kw = _preprocess_event(**kw)
    logargs, logkwargs = _parse_job_event(**kw)
    _send_event(*logargs, **logkwargs)


def send_workflow_event(**kw):
    kw = _preprocess_event(**kw)
    logargs, logkwargs = _parse_workflow_event(**kw)
    _send_event(*logargs, **logkwargs)


def send_task_event(**kw):
    kw = _preprocess_event(**kw)
    logargs, logkwargs = _parse_task_event(**kw)
    _send_event(*logargs, **logkwargs)


def timestamp() -> str:
    return datetime.now().astimezone().isoformat()


def _preprocess_event(execinfo: ExecInfoType = None, **logkwargs):
    if execinfo:
        logkwargs = {**execinfo, **logkwargs}
    if not logkwargs.get("time"):
        logkwargs["time"] = timestamp()
    _parse_exception(logkwargs)
    return logkwargs


def _parse_exception(logkwargs):
    exception = logkwargs.pop("exception", None)
    if exception is None:
        return
    elif isinstance(exception, str):
        logkwargs["error"] = True
        tb = exception
        error_message = _extract_reason_from_traceback(tb)
        if not logkwargs.get("error_message"):
            logkwargs["error_message"] = error_message
        if not logkwargs.get("error_traceback"):
            logkwargs["error_traceback"] = tb
    elif isinstance(exception, BaseException):
        logkwargs["error"] = True
        tb = "".join(
            traceback.format_exception(
                type(exception), exception, exception.__traceback__
            )
        )
        if not logkwargs.get("error_traceback"):
            logkwargs["error_traceback"] = tb
        if logkwargs.get("context") == "node":
            exception = _first_exception(exception)
        if not logkwargs.get("error_message"):
            error_message = str(exception)
            if not error_message:
                error_message = type(exception).__name__
            logkwargs["error_message"] = error_message
    else:
        raise TypeError(
            "ewoks event field 'exception' should be a string or an exception instance"
        )


def _first_exception(exception: BaseException) -> BaseException:
    while True:
        if exception.__cause__ is not None:
            # exception raised from another exception
            exception = exception.__cause__
        elif exception.__context__ is not None:
            # exception raised during handling of another exception
            exception = exception.__context__
        else:
            return exception


def _extract_reason_from_traceback(tb: str) -> str:
    for line in reversed(tb.split("\n")):
        if ":" in line:
            return line
    return ""


def _parse_workflow_event(**logkwargs):
    event_data, logkwargs = _extract_workflow_fields(**logkwargs)
    msg = _validate_event(event_data)
    logargs = (
        "[job %r] [workflow %r] %s",
        event_data["job_id"],
        event_data["workflow_id"],
        msg,
    )
    logkwargs["event_data"] = event_data
    return logargs, logkwargs


def _parse_job_event(**logkwargs):
    event_data, logkwargs = _extract_job_fields(**logkwargs)
    msg = _validate_event(event_data)
    logargs = ("[job %r] %s", event_data["job_id"], msg)
    logkwargs["event_data"] = event_data
    return logargs, logkwargs


def _parse_task_event(**logkwargs):
    event_data, logkwargs = _extract_task_fields(**logkwargs)
    msg = _validate_event(event_data)
    logargs = (
        "[job %r] [workflow %r] [node %r] [task %r] %s",
        event_data["job_id"],
        event_data["workflow_id"],
        event_data["node_id"],
        event_data["task_id"],
        msg,
    )
    logkwargs["event_data"] = event_data
    return logargs, logkwargs


def _extract_common_fields(
    host_name: str,
    process_id: int,
    user_name: str,
    job_id: str,
    event: str,
    engine: Optional[str] = None,
    time: Optional[str] = None,
    error: Optional[bool] = None,
    error_message: Optional[str] = None,
    error_traceback: Optional[str] = None,
    **logkwargs,
):
    event_data = {
        "host_name": host_name,
        "process_id": process_id,
        "user_name": user_name,
        "job_id": job_id,
        "engine": engine,
        "type": event,
        "time": time,
        "error": error,
        "error_message": error_message,
        "error_traceback": error_traceback,
    }
    return event_data, logkwargs


def _extract_job_fields(
    **logkwargs,
):
    event_data, logkwargs = _extract_common_fields(**logkwargs)
    event_data["context"] = "job"
    return event_data, logkwargs


def _extract_workflow_fields(
    workflow_id: str,
    **logkwargs,
):
    event_data, logkwargs = _extract_job_fields(**logkwargs)
    event_data["context"] = "workflow"
    event_data["workflow_id"] = workflow_id
    return event_data, logkwargs


def _extract_task_fields(
    node_id: str,
    task_id: str,
    progress: Optional[Number] = None,
    task_uri: Optional[str] = None,
    input_uris: Optional[List[Dict[str, Optional[str]]]] = None,
    output_uris: Optional[List[Dict[str, Optional[str]]]] = None,
    **logkwargs,
):
    event_data, logkwargs = _extract_workflow_fields(**logkwargs)
    event_data["context"] = "node"
    event_data["node_id"] = node_id
    event_data["task_id"] = task_id
    event_data["progress"] = progress
    event_data["task_uri"] = task_uri
    event_data["input_uris"] = input_uris
    event_data["output_uris"] = output_uris
    return event_data, logkwargs


def _validate_event(event_data: dict) -> str:
    event_data.update((field, None) for field in set(FIELD_TYPES) - set(event_data))
    event = event_data["type"]
    if event_data["context"] == "job":
        obj = "job"
    elif event_data["context"] == "workflow":
        obj = "workflow"
    elif event_data["context"] == "node":
        obj = "task"
    else:
        raise ValueError(f"context '{event_data['context']}' is unknown")
    if event == "start":
        return f"{obj} started"
    elif event == "end":
        error = bool(
            event_data["error"]
            or event_data["error_message"]
            or event_data["error_traceback"]
        )
        event_data["error"] = error
        if error:
            if event_data["error_message"] is None:
                event_data["error_message"] = ""
            if event_data["error_traceback"] is None:
                event_data["error_traceback"] = ""
        if not error:
            return f"{obj} finished"
        elif event_data["error"] and event_data["error_message"]:
            return f"{obj} failed ({event_data['error_message']})"
        else:
            return f"{obj} failed"
    elif event == "progress":
        if not isinstance(event_data["progress"], Number):
            event_data["progress"] = float("nan")
        return f"{obj} progress {event_data['progress']}%"
    else:
        raise ValueError(f"unknown ewoks event type '{event}'")


def _send_event(msg: str, *args, event_data=None, extra=None, contexts=None, **kw):
    if not isinstance(event_data, Mapping):
        raise TypeError("'event_data' should be a mapping")
    if extra:
        if not isinstance(extra, Mapping):
            raise TypeError("'extra' should be a mapping")
        extra = {**extra, **event_data}
    else:
        extra = event_data
    global_state.send(msg, *args, extra=extra, **kw)
